from __future__ import annotations

from abc import ABC
from typing import Any, Callable, Optional, Sequence, Tuple, Union

from ._types import JSONSchema
from .operations import (
    SEARCH_OP_LT,
    SEARCH_OP_LE,
    SEARCH_OP_EQ,
    SEARCH_OP_GT,
    SEARCH_OP_GE,
    SEARCH_OP_IN,
    SEARCH_OP_CO,
    SEARCH_OP_ICO,
    SEARCH_OP_ISW,
    SEARCH_OP_IEW,
    SEARCH_OP_LIKE,
    SEARCH_OP_ILIKE,
)


__all__ = [
    "FUNCTION_AND",
    "FUNCTION_OR",
    "FUNCTION_NOT",
    "FUNCTION_LT",
    "FUNCTION_LE",
    "FUNCTION_EQ",
    "FUNCTION_GT",
    "FUNCTION_GE",
    "FUNCTION_CO",
    "FUNCTION_ICO",
    "FUNCTION_IN",
    "FUNCTION_RESOLVE",
    "FUNCTION_LIST",
    "FUNCTION_HELPER_WC",

    "VALID_FUNCTIONS",

    "FUNCTION_SEARCH_OP_MAP",

    "LiteralValue",
    "FunctionName",
    "Query",
    "AST",
    "Args",

    "Expression",
    "Literal",

    "convert_query_to_ast",
    "convert_query_to_ast_and_preprocess",
    "ast_to_and_asts",
    "and_asts_to_ast",
    "check_operation_permissions",
]


FUNCTION_AND = "#and"
FUNCTION_OR = "#or"
FUNCTION_NOT = "#not"

FUNCTION_LT = "#lt"
FUNCTION_LE = "#le"
FUNCTION_EQ = "#eq"
FUNCTION_GT = "#gt"
FUNCTION_GE = "#ge"

FUNCTION_IN = "#in"

FUNCTION_CO = "#co"  # contains
FUNCTION_ICO = "#ico"  # case-insensitive string contains

FUNCTION_ISW = "#isw"  # case-insensitive starts with
FUNCTION_IEW = "#iew"  # case-insensitive ends with
FUNCTION_LIKE = "#like"  # SQL-style LIKE operator
FUNCTION_ILIKE = "#ilike"  # SQL-style case-insensitive ILIKE operator

FUNCTION_RESOLVE = "#resolve"
FUNCTION_LIST = "#list"

FUNCTION_HELPER_WC = "#_wc"

# Special function which is just used for Python typing to represent a literal.
FUNCTION_HELPER_LITERAL = "#_literal"

VALID_FUNCTIONS = (
    FUNCTION_AND,
    FUNCTION_OR,
    FUNCTION_NOT,
    FUNCTION_LT,
    FUNCTION_LE,
    FUNCTION_EQ,
    FUNCTION_GT,
    FUNCTION_GE,
    FUNCTION_IN,
    FUNCTION_CO,
    FUNCTION_ICO,
    FUNCTION_ISW,
    FUNCTION_IEW,
    FUNCTION_LIKE,
    FUNCTION_ILIKE,
    FUNCTION_RESOLVE,
    FUNCTION_LIST,
    FUNCTION_HELPER_WC,
)

BINARY_RANGE = (2, 2)

# Keys are functions, values are a tuple of (minimum argument count, maximum argument count)
FUNCTION_ARGUMENTS = {
    FUNCTION_AND: BINARY_RANGE,
    FUNCTION_OR: BINARY_RANGE,
    FUNCTION_NOT: (1, 1),
    FUNCTION_LT: BINARY_RANGE,
    FUNCTION_LE: BINARY_RANGE,
    FUNCTION_EQ: BINARY_RANGE,
    FUNCTION_GT: BINARY_RANGE,
    FUNCTION_GE: BINARY_RANGE,
    FUNCTION_CO: BINARY_RANGE,
    FUNCTION_ICO: BINARY_RANGE,
    FUNCTION_IN: BINARY_RANGE,
    FUNCTION_ISW: BINARY_RANGE,
    FUNCTION_IEW: BINARY_RANGE,
    FUNCTION_LIKE: BINARY_RANGE,
    FUNCTION_ILIKE: BINARY_RANGE,
    FUNCTION_RESOLVE: (0, None),
    FUNCTION_LIST: (1, None),
    FUNCTION_HELPER_WC: BINARY_RANGE,
}


FUNCTION_SEARCH_OP_MAP = {
    FUNCTION_LT: SEARCH_OP_LT,
    FUNCTION_LE: SEARCH_OP_LE,
    FUNCTION_EQ: SEARCH_OP_EQ,
    FUNCTION_GT: SEARCH_OP_GT,
    FUNCTION_GE: SEARCH_OP_GE,

    FUNCTION_CO: SEARCH_OP_CO,
    FUNCTION_ICO: SEARCH_OP_ICO,

    FUNCTION_IN: SEARCH_OP_IN,

    FUNCTION_ISW: SEARCH_OP_ISW,
    FUNCTION_IEW: SEARCH_OP_IEW,

    FUNCTION_LIKE: SEARCH_OP_LIKE,
    FUNCTION_ILIKE: SEARCH_OP_ILIKE,
}

# literal types are scalar that are used in queries to compare the value for a
# given field for example 20, in a query for age greater than 20.
# In the AST they are distinct from Expressions
literal_types = str, int, float, bool    # TODO: How to handle dict in practical cases?
LiteralValue = Union[str, int, float, bool]

FunctionName = str


Query = Union[list, LiteralValue]


# TODO: Prevent nested resolves
# TODO: Argument types...


# AST = Union["Expression", "Literal"]
Args = Tuple["AST", ...]


class AST(ABC):
    type: str
    fn: str
    args: Args
    value: Any


class Expression(AST):
    type = "e"

    def __init__(self, fn: str, args: Sequence[AST]):
        assert fn in VALID_FUNCTIONS
        self.fn = fn

        arg_range = FUNCTION_ARGUMENTS[fn]
        assert len(args) >= arg_range[0] and (arg_range[1] is None or len(args) <= arg_range[1])
        self.args: Args = tuple(args)

    def __eq__(self, other):
        return (isinstance(other, Expression) and self.fn == other.fn and
                all(a == b for a, b in zip(self.args, other.args)))

    @property
    def value(self) -> Expression:
        return self

    def __str__(self):
        return "[{}, {}]".format(self.fn, ", ".join(str(a) for a in self.args))

    def __repr__(self):  # pragma: no cover
        return f"<Expression {self.fn} [{', '.join(repr(a) for a in self.args)}]>"


class Literal(AST):
    type = "l"
    fn = FUNCTION_HELPER_LITERAL
    args = ()

    def __init__(self, value: LiteralValue):
        assert any(isinstance(value, t) for t in literal_types)
        self.value = value

    def __eq__(self, other):
        return isinstance(other, Literal) and self.value == other.value

    def __str__(self):
        return str(self.value)

    def __repr__(self):  # pragma: no cover
        return f"<Literal {self.value}>"

    def __hash__(self):
        return self.value.__hash__()


def convert_query_to_ast(query: Query) -> AST:
    """
    Recursive function that converts a list of nested lists and scalars to
    an AST object containing nested Expression and Literal objects.
    In the Query argument, a list defines an expression and otherwise the
    scalar values should define the literals.
    For example:
    ["#eq", ["#resolve", "patient", "name"], "John Doe"]
    |<---------------    Root List 1   ---------------->
    |<fn>|  |<-------    arg #1   ------->|  |<-arg #2>|
            |<----    Nested List 2  ---->|, |<literal>|
            |<- fn ->|   |<arg 1>| |<arg2>|
    is converted into
    Expression -->
        - type: "e"
        - fn: FUNCTION_EQ,
        - range: (2, 2)
        - args: Tuple(
            Expression -->
                - type: "e"
                - fn: FUNCTION_RESOLVE
                - range: (1, None)
                - args: Tuple("patient", "name")
                        Note: Does not recurse down to literals due to function #resolve implementation)
            Literal -->
                - type: "l"
                - value: "John Doe"
        )
    """
    if isinstance(query, list):
        if len(query) == 0 or not isinstance(query[0], str) or query[0] not in VALID_FUNCTIONS:
            raise SyntaxError("Invalid expression: {}".format(query))

        try:
            return Expression(query[0], tuple(convert_query_to_ast(q) for q in query[1:]))
        except AssertionError:
            raise SyntaxError("Invalid number of arguments for function {}: {}".format(query[0], len(query[1:])))

    elif any(isinstance(query, t) for t in literal_types):
        return Literal(query)

    raise ValueError("Invalid literal: {}".format(query))


def simplify_nots(ast: AST) -> AST:
    # not (not a) => a

    if ast.type == "l":
        return ast

    if ast.fn == FUNCTION_NOT and ast.args[0].type == "e" and ast.args[0].fn == FUNCTION_NOT:
        return simplify_nots(ast.args[0].args[0])

    return Expression(fn=ast.fn, args=[simplify_nots(a) for a in ast.args])


def convert_query_to_ast_and_preprocess(query: Query) -> AST:
    ast = convert_query_to_ast(query)
    return simplify_nots(ast)


def ast_to_and_asts(ast: AST) -> Tuple[AST, ...]:
    # (and e1 e2) => <e1, e2>
    # (and (and e1 e2) e3) => <e1, e2, e3>
    # (and e1 (and e2 e3)) => <e1, e2, e3>
    # (and (and e1 e2) (and e3 e4)) => <e1, e2, e3, e4>
    # etc.

    if not ast.type == "e" or ast.fn != FUNCTION_AND:
        return ast,

    return (*ast_to_and_asts(ast.args[0]), *ast_to_and_asts(ast.args[1]))


def _and_asts_to_ast_rec(asts: Tuple[AST, ...]) -> AST:
    if len(asts) == 1:  # Base case
        return asts[0]

    return Expression(FUNCTION_AND, [asts[0], _and_asts_to_ast_rec(asts[1:])])


def and_asts_to_ast(asts: Tuple[AST, ...]) -> Optional[AST]:
    # ()               => None
    # (e1, e2, e3, e4) => (and e1 (and e2 (and e3 e4)))

    if len(asts) == 0:
        return None

    return _and_asts_to_ast_rec(asts)


def check_operation_permissions(
    ast: AST,
    schema: JSONSchema,
    search_getter: Callable[[Tuple[Literal, ...], dict], dict],
    internal: bool = False
):
    if ast.type == "l":
        return

    if ast.fn == FUNCTION_RESOLVE:
        search_properties = search_getter(ast.args, schema)

        query_modes = ("internal", "all") if internal else ("all",)
        query_mode = search_properties.get("queryable", "none")

        # If resolving any field, make sure at least some operation is permitted
        if query_mode not in query_modes:
            # TODO: Custom exception?
            raise ValueError(f"Cannot access field using {ast} (queryable: {query_mode}, allowed are {query_modes})")

    # Check to make sure the function's execution is permitted

    if ast.fn not in FUNCTION_SEARCH_OP_MAP:
        return

    # TODO: Make this check recursive (#11) or somehow deal with boolean values
    if any(FUNCTION_SEARCH_OP_MAP[ast.fn] not in search_getter(a.args, schema).get("operations", [])
           for a in ast.args if a.type == "e" and a.fn == FUNCTION_RESOLVE):
        # TODO: Custom exception?
        raise ValueError(f"Schema forbids using function: {ast.fn}\nAST: {ast}\nSchema: \n{schema}")



from datetime import datetime
from typing import Dict, List, Tuple, Union

from . import data_structure
from . import operations
from . import postgres
from . import queries

__all__ = [
    "build_search_response",

    "data_structure",
    "operations",
    "postgres",
    "queries",
]


def build_search_response(results: Union[Dict, List, Tuple], start_time: datetime) -> Dict:
    return {
        "results": results,
        "time": (datetime.now() - start_time).total_seconds()
    }



import logging

from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.http import JsonResponse, HttpRequest, HttpResponse
from rest_framework.request import Request as DrfRequest
from typing import Awaitable, Callable

from bento_lib.responses.errors import http_error
from bento_lib.auth.exceptions import BentoAuthException
from bento_lib.auth.middleware.base import BaseAuthMiddleware

__all__ = [
    "DjangoAuthMiddleware",
]


class DjangoAuthMiddleware(BaseAuthMiddleware):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # If no logger was passed, create a new logger
        if self._logger is None:
            self._logger = logging.getLogger(__name__)

    def get_authz_header_value(self, request: DrfRequest | HttpRequest) -> str | None:
        return request.headers.get("Authorization")

    @staticmethod
    def mark_authz_done(request: DrfRequest | HttpRequest):
        # noinspection PyProtectedMember
        req = request._request if isinstance(request, DrfRequest) else request
        req.bento_determined_authz = True

    def make_django_middleware(self):
        # noinspection PyMethodParameters
        class InnerMiddleware:
            async_capable = True
            sync_capable = False

            def __init__(inner_self, get_response: Callable[[HttpRequest], Awaitable[HttpResponse]]):
                inner_self.get_response = get_response
                if iscoroutinefunction(inner_self.get_response):  # pragma: no cover
                    markcoroutinefunction(inner_self)

            async def __call__(inner_self, request: HttpRequest) -> HttpResponse:
                return await self.dispatch(inner_self.get_response, request)

            @staticmethod
            def process_exception(request: HttpRequest, exc: Exception):
                if isinstance(exc, BentoAuthException):
                    self.mark_authz_done(request)
                    return self._make_auth_error(exc)
                return None

        return InnerMiddleware

    def _make_auth_error(self, e: BentoAuthException) -> JsonResponse:
        return JsonResponse(
            http_error(
                e.status_code,
                e.message,
                drs_compat=self._drs_compat,
                sr_compat=self._sr_compat,
                beacon_meta_callback=self._beacon_meta_callback,
            ),
            status=e.status_code)

    async def dispatch(
        self,
        get_response: Callable[[HttpRequest], Awaitable[HttpResponse]],
        request: HttpRequest,
    ) -> HttpResponse:
        if not self.enabled:
            return await get_response(request)

        request.bento_determined_authz = False

        # Don't handle BentoAuthException here - middleware handles it elsewhere
        response = await get_response(request)  # We've just crammed a new property in there... no state object

        if not self.enabled or self.request_is_exempt(request.method, request.path):
            # - Skip checks if the authorization middleware is disabled
            # - Allow pre-flight responses through, as well as any configured exempt URLs
            return response

        if not request.bento_determined_authz:
            # Next in response chain didn't properly think about auth; return 403
            return self._make_auth_error(BentoAuthException("Forbidden", status_code=403))

        # Otherwise, return the response as normal
        return response



from abc import abstractmethod
from typing import Any


__all__ = ["MarkAuthzDoneMixin"]


class MarkAuthzDoneMixin:
    @staticmethod
    @abstractmethod
    def mark_authz_done(request: Any):  # pragma: no cover
        pass



import logging
from typing import Literal

__all__ = [
    "LogLevelLiteral",
    "log_level_from_str",
]


LogLevelLiteral = Literal["debug", "info", "warning", "error"]


log_level_str_to_log_level: dict[str, int] = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


def log_level_from_str(level: str, default: int = logging.INFO) -> int:
    return log_level_str_to_log_level.get(level.lower(), default)



import json

from bento_lib.logging import LogLevelLiteral
from bento_lib.service_info.constants import SERVICE_ORGANIZATION_C3G_PYDANTIC
from bento_lib.service_info.types import GA4GHServiceOrganizationModel
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings, EnvSettingsSource, PydanticBaseSettingsSource, SettingsConfigDict
from typing import Any

__all__ = [
    "CorsOriginsParsingEnvSettingsSource",
    "BentoBaseConfig",
    "BentoFastAPIBaseConfig",
]


CORS_ORIGINS_DEFAULT: tuple[str, ...] = ()


class CorsOriginsParsingEnvSettingsSource(EnvSettingsSource):
    def prepare_field_value(self, field_name: str, field: FieldInfo, value: Any, value_is_complex: bool) -> Any:
        if field_name == "cors_origins":
            return tuple(x.strip() for x in value.split(";")) if value is not None else CORS_ORIGINS_DEFAULT
        return json.loads(value) if value_is_complex else value


class BentoBaseConfig(BaseSettings):
    bento_debug: bool = False
    bento_container_local: bool = False
    bento_validate_ssl: bool = True
    bento_authz_enabled: bool = True
    bento_authz_service_url: str  # Bento authorization service base URL

    service_id: str
    service_name: str
    service_description: str = ""  # If description is blank, it should be stripped out in the response
    service_url_base_path: str = "http://127.0.0.1:5000"  # Base path to construct URIs from
    service_contact_url: str = "mailto:info@c3g.ca"
    service_organization: GA4GHServiceOrganizationModel = SERVICE_ORGANIZATION_C3G_PYDANTIC

    log_level: LogLevelLiteral = "debug"

    cors_origins: tuple[str, ...] = CORS_ORIGINS_DEFAULT

    # Make Config instances hashable + immutable
    model_config = SettingsConfigDict(frozen=True)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            CorsOriginsParsingEnvSettingsSource(settings_cls),
            dotenv_settings,
            file_secret_settings,
        )


class BentoFastAPIBaseConfig(BentoBaseConfig):
    service_docs_path: str = "/docs"
    service_openapi_path: str = "/openapi.json"



# !!! LEGACY FILE !!!

__all__ = [
    "BENTO_USER_HEADER",
    "BENTO_USER_ROLE_HEADER",

    "DJANGO_USER_HEADER",
    "DJANGO_USER_ROLE_HEADER",
]


def _to_django_header(header: str):
    return f"HTTP_{header.replace('-', '_').upper()}"


BENTO_USER_HEADER = "X-User"
BENTO_USER_ROLE_HEADER = "X-User-Role"

DJANGO_USER_HEADER = _to_django_header(BENTO_USER_HEADER)
DJANGO_USER_ROLE_HEADER = _to_django_header(BENTO_USER_ROLE_HEADER)



from .permissions import Permission, LEVEL_PROJECT, LEVEL_DATASET, PERMISSIONS

__all__ = [
    "permission_valid_for_resource",
    "valid_permissions_for_resource",
]


def permission_valid_for_resource(permission: Permission, resource: dict) -> bool:
    valid_data_type_narrowing = permission.supports_data_type_narrowing or "data_type" not in resource

    if permission.min_level_required == LEVEL_DATASET:
        return valid_data_type_narrowing
    elif permission.min_level_required == LEVEL_PROJECT:
        return valid_data_type_narrowing and "dataset" not in resource
    else:  # LEVEL_INSTANCE
        return (
            valid_data_type_narrowing
            and "dataset" not in resource
            and "project" not in resource
            and resource.get("everything", False)
        )
        # otherwise, invalid resource (so False)


def valid_permissions_for_resource(resource: dict) -> list[Permission]:
    return [p for p in PERMISSIONS if permission_valid_for_resource(p, resource)]



from typing import Literal
from .types import GA4GHServiceOrganization, GA4GHServiceOrganizationModel

__all__ = [
    "SERVICE_ENVIRONMENT_DEV",
    "SERVICE_ENVIRONMENT_PROD",
    "SERVICE_ORGANIZATION_C3G",
    "SERVICE_ORGANIZATION_C3G_PYDANTIC",
    "SERVICE_GROUP_BENTO",
]


SERVICE_ENVIRONMENT_DEV: Literal["dev"] = "dev"
SERVICE_ENVIRONMENT_PROD: Literal["prod"] = "prod"

SERVICE_ORGANIZATION_C3G: GA4GHServiceOrganization = {"name": "C3G", "url": "https://www.computationalgenomics.ca"}
SERVICE_ORGANIZATION_C3G_PYDANTIC: GA4GHServiceOrganizationModel = GA4GHServiceOrganizationModel.model_validate(
    SERVICE_ORGANIZATION_C3G)

SERVICE_GROUP_BENTO = "ca.c3g.bento"



from datetime import datetime, timezone
from functools import partial
from werkzeug.http import HTTP_STATUS_CODES

from typing import Callable


__all__ = [
    "http_error",

    "bad_request_error",
    "unauthorized_error",
    "forbidden_error",
    "not_found_error",
    "method_not_allowed_error",
    "not_acceptable_error",
    "request_timeout_error",
    "range_not_satisfiable_error",

    "internal_server_error",
    "not_implemented_error",
]


def _error_message(message):
    return {"message": message}


def http_error(
    code: int,
    *errors,
    drs_compat: bool = False,
    sr_compat: bool = False,
    beacon_meta_callback: Callable[[], dict] | None = None,
):
    """
    Builds a dictionary for an HTTP error JSON response.
    :param code: The error status code to embed in the response.
    :param errors: A list of error descriptions (human-readable) to explain the error.
    :param drs_compat: Whether to generate a GA4GH DRS schema backwards-compatible response.
    :param sr_compat: Whether to generate a GA4GH Service Registry backwards-compatible response.
    :param beacon_meta_callback: Callback for generating GA4GH Beacon V2 backwards-compatible meta field for
           error response. If this is specified, Beacon V2-compatible errors will be enabled.
    :return: A dictionary to encode in JSON for the error response.
    """

    if code not in HTTP_STATUS_CODES:
        print(f"[Bento Lib] Error: Could not find code {code} in valid HTTP status codes.")
        code = 500
        errors = (*errors, f"An invalid status code of {code} was specified by the service.")

    if code < 400:
        print(f"[Bento Lib] Error: Code {code} is not an HTTP error code.")
        code = 500
        errors = (*errors, f"A non-error status code of {code} was specified by the service.")

    message = HTTP_STATUS_CODES[code]

    return {
        "code": code,
        "message": message,
        "timestamp": datetime.now(timezone.utc).isoformat("T").split("+")[0] + "Z",
        **({"errors": [_error_message(e) for e in errors]} if errors else {}),

        # The DRS spec has a slightly different error specification - if a
        # compatibility flag is passed in, extra fields are tacked on here.
        **({"status_code": code, "msg": message} if drs_compat else {}),

        # The Service Registry spec *also* has a slightly different error
        # specification; do the same thing as above.
        **({
            "status": code,
            "title": message,
            **({"detail": " | ".join(errors)} if errors else {}),
        } if sr_compat else {}),

        # ... why so many "standards"? Here's a Beacon V2-compatible error specification
        **({
            "meta": beacon_meta_callback(),
            "error": {
                "errorCode": code,
                **({"errorMessage": " | ".join(errors)} if errors else {}),
            },
        } if beacon_meta_callback is not None else {}),
    }


_e = partial(partial, http_error)

bad_request_error = _e(400)
unauthorized_error = _e(401)
forbidden_error = _e(403)
not_found_error = _e(404)
method_not_allowed_error = _e(405)
not_acceptable_error = _e(406)
request_timeout_error = _e(408)
range_not_satisfiable_error = _e(416)

internal_server_error = _e(500)
not_implemented_error = _e(501)



__all__ = [
    "SEARCH_OP_EQ",
    "SEARCH_OP_LT",
    "SEARCH_OP_LE",
    "SEARCH_OP_GT",
    "SEARCH_OP_GE",

    "SEARCH_OP_CO",
    "SEARCH_OP_ICO",

    "SEARCH_OP_IN",

    "SEARCH_OP_ISW",
    "SEARCH_OP_IEW",
    "SEARCH_OP_LIKE",
    "SEARCH_OP_ILIKE",

    "SEARCH_OPERATIONS",
]


SEARCH_OP_EQ = "eq"
SEARCH_OP_LT = "lt"
SEARCH_OP_LE = "le"
SEARCH_OP_GT = "gt"
SEARCH_OP_GE = "ge"

SEARCH_OP_CO = "co"
SEARCH_OP_ICO = "ico"

SEARCH_OP_IN = "in"

SEARCH_OP_ISW = "isw"
SEARCH_OP_IEW = "iew"
SEARCH_OP_LIKE = "like"
SEARCH_OP_ILIKE = "ilike"

SEARCH_OPERATIONS = (
    SEARCH_OP_EQ,
    SEARCH_OP_LT,
    SEARCH_OP_LE,
    SEARCH_OP_GT,
    SEARCH_OP_GE,
    SEARCH_OP_IN,
    SEARCH_OP_CO,
    SEARCH_OP_ICO,
    SEARCH_OP_ISW,
    SEARCH_OP_IEW,
    SEARCH_OP_LIKE,
    SEARCH_OP_ILIKE,
    SEARCH_OP_ILIKE,
)



import functools
import re

from psycopg2 import sql
from typing import Callable, Dict, Literal, Optional, Tuple

from . import queries as q
from ._types import JSONSchema


# Search Rules:
#  - If an object or query doesn't match the schema, it's an error.
#  - If an optional property isn't present, it's "False".


__all__ = ["search_query_to_psycopg2_sql"]


SQLComposableWithParams = Tuple[sql.Composable, tuple]

QUERY_ROOT = q.Literal("$root")
SQL_ROOT = sql.Identifier("_root")
SQL_NOTHING = sql.SQL("")


# TODO: Python 3.7: Data Class
class OptionalComposablePair:
    def __init__(self, parent: Optional[sql.Composable], current: Optional[sql.Composable]):
        # A 'Composable' here is an abstract base class from PsycoPG2's Postgres IR which can represent anything
        # in the PostgreSQL language which can be composed together to create a query.
        self.parent: Optional[sql.Composable] = parent
        self.current: Optional[sql.Composable] = current

    def __repr__(self):  # pragma: no cover
        return f"<OptionalComposablePair parent={self.parent} current={self.current}>"


# TODO: Python 3.7: Data Class
class JoinAndSelectData:
    def __init__(
        self,
        relations: OptionalComposablePair,
        aliases: OptionalComposablePair,
        current_alias_str: Optional[str],
        current_alias_sql_schema: Optional[sql.SQL],
        key_link: Optional[Tuple[str, str]],
        field_alias: Optional[str],
        search_properties: dict,
        unresolved: Tuple[q.Literal, ...]
    ):
        self.relations: OptionalComposablePair = relations
        self.aliases: OptionalComposablePair = aliases
        self.current_alias_str: Optional[str] = current_alias_str
        self.current_alias_sql_schema: Optional[sql.SQL] = current_alias_sql_schema
        self.key_link: Optional[Tuple[str, str]] = key_link
        self.field_alias: Optional[str] = field_alias
        self.search_properties: dict = search_properties
        self.unresolved: Tuple[q.Literal, ...] = unresolved

    def __repr__(self):  # pragma: no cover
        return f"<JoinAndSelectData relations={self.relations} aliases={self.aliases}>"


def json_schema_to_postgres_type(schema: JSONSchema, structure_type: Literal["json", "jsonb"]) -> str:
    """
    Maps a JSON schema to a Postgres type for on the fly mapping.
    :param schema: JSON schema to map.
    :param structure_type: Whether we are deconstructing a JSON or JSONB field.
    """
    if schema["type"] == "string":
        return "TEXT"
    elif schema["type"] == "integer":
        return "INTEGER"
    elif schema["type"] == "number":
        return "DOUBLE PRECISION"
    elif schema["type"] == "object":
        return structure_type.upper()
    elif schema["type"] == "array":
        return structure_type.upper()
    elif schema["type"] == "boolean":
        return "BOOLEAN"
    else:
        # null
        return "TEXT"  # TODO


def json_schema_to_postgres_schema(
    name: str,
    schema: JSONSchema,
    structure_type: Literal["json", "jsonb"],
) -> Tuple[Optional[sql.Composable], Optional[str], Optional[sql.Composable]]:
    """
    Maps a JSON object schema to a Postgres schema for on-the-fly mapping.
    :param name: the name to give the fake table.
    :param schema: JSON schema to map.
    :param structure_type: Whether we are deconstructing a JSON or JSONB field.
    """

    if schema["type"] != "object":
        return None, None, None

    return (
        sql.Identifier(name),
        name,
        sql.SQL("({})").format(
            sql.SQL(", ").join(
                sql.SQL("{} {}").format(
                    sql.Identifier(p),
                    sql.SQL(json_schema_to_postgres_type(s, structure_type)),
                )
                for p, s in schema["properties"].items()
            )
        ),
    )


def _get_search_and_database_properties(schema: JSONSchema) -> Tuple[dict, dict]:
    search_properties = schema.get("search", {})
    return search_properties, search_properties.get("database", {})


def collect_resolve_join_tables(
    resolve: Tuple[q.Literal, ...],
    schema: JSONSchema,
    parent_relation: Optional[Tuple[Optional[sql.Composable], Optional[sql.Composable]]] = None,
    aliased_resolve_path: Optional[str] = None
) -> Tuple[JoinAndSelectData, ...]:
    """
    Recursively collects tables to join for compiling the query.
    :param resolve: The current resolve list, minus the command. Starts with $root to keep schema proper.
    :param schema: The schema of the current element (first property.)
    :param parent_relation: The tuple representing the parent relation, format: (relation name, alias).
    :param aliased_resolve_path: Underscore-delimited string 'list' of parent property names in the case of nested
                                 objects/arrays, for building aliases for the nested fields in question.
                                 If it is None, we are at the root.
    :return: Tuple of tables with joining properties.
    """

    # Many to one: child_key -> primary_key
    # Many to many: primary_key -> parent_key (treat as just another through)

    if len(resolve) == 0:
        return ()

    search_properties, search_database_properties = _get_search_and_database_properties(schema)

    # TODO: might be able to inject .value under some circumstances here
    schema_field = resolve[0].value
    db_field = search_database_properties.get("field", schema_field if schema_field != QUERY_ROOT.value else None)
    current_relation = search_database_properties.get("relation")

    # Gradually (recursively) build up an alias for this resolve statement at the field we're currently working on.
    # For example, if we are accessing $root.biosamples.[item].id, this would get the alias:
    #  _root_biosamples_item_id
    # by the process of:
    #  - replacing the $ in $root with _
    #  - recursing down and combining the path with the field as it is encountered using '_'
    aliased_resolve_path_str: str = aliased_resolve_path or ""
    new_aliased_resolve_path = re.sub(r"[$\[\]]+", "", f"{aliased_resolve_path_str}_{schema_field}")
    current_alias = None
    current_alias_str = None
    current_alias_sql_schema = None

    if current_relation is None and schema["type"] in ("array", "object"):
        if db_field is None:
            raise SyntaxError("Cannot determine or synthesize current relation")

        # TODO: Additional conditions to check if it's actually JSON
        # TODO: HStore support?
        structure_type = search_database_properties.get("type")

        if structure_type in ("json", "jsonb"):
            # For drilling into Postgres JSON objects, we have to be a bit more clever about how we access/alias
            # the fields in question...
            if schema["type"] == "array":  # JSON(B) array or object
                # will be used to call either json_array_elements(...) or jsonb_array_elements(...):
                relation_sql_template = "{structure_type}_array_elements({field})"
                current_alias = sql.Identifier(new_aliased_resolve_path)
                current_alias_str = new_aliased_resolve_path
            else:  # object
                # will be used to call either json_to_record(...) or jsonb_to_record(...):
                relation_sql_template = "{structure_type}_to_record({field})"
                current_alias, current_alias_str, current_alias_sql_schema = json_schema_to_postgres_schema(
                    new_aliased_resolve_path, schema, structure_type)

            current_relation = sql.SQL(relation_sql_template).format(
                structure_type=sql.SQL(structure_type),  # json or jsonb here
                field=sql.SQL(".").join(
                    # TODO: Write comment about logic here - array vs obj
                    (parent_relation[1], sql.Identifier(db_field))
                    if db_field != "[item]" else (parent_relation[1],)
                ))

        elif structure_type == "array" and schema["type"] == "array":  # Postgres array
            # TODO: document what unnest is up to
            current_relation = sql.SQL("unnest({relation}.{field})").format(
                relation=parent_relation[1],
                field=sql.Identifier(db_field),
            )
            current_alias = sql.Identifier(new_aliased_resolve_path)
            current_alias_str = new_aliased_resolve_path

        else:
            raise ValueError(
                f"Structure type / schema type mismatch: {structure_type} / {schema['type']}\n"
                f"    Search properties: {search_properties}\n"
                f"    Search database properties: {search_database_properties}\n"
                f"    Aliased resolve path: {new_aliased_resolve_path}"
            )

    elif current_relation is not None:
        current_relation = sql.Identifier(current_relation)
        current_alias = sql.Identifier(new_aliased_resolve_path)
        current_alias_str = new_aliased_resolve_path

    relations = OptionalComposablePair(parent_relation[0] if parent_relation is not None else None, current_relation)

    # Parent, Current
    aliases = OptionalComposablePair(
        parent=sql.Identifier(aliased_resolve_path) if aliased_resolve_path is not None else None,
        current=current_alias)

    key_link = None
    if "relationship" in search_database_properties:
        relationship = search_database_properties["relationship"]
        relationship_type = relationship["type"]
        # TODO: py3.10: match
        if relationship_type == "MANY_TO_ONE":
            # For a many-to-one relationship, the outer/higher element in the nesting has the foreign key,
            # so that many outer elements can map to the same inner element.
            key_link = (relationship["foreign_key"], search_database_properties["primary_key"])
        elif relationship_type == "ONE_TO_MANY":
            # For a one-to-many relationship, the inner/nested element in the relationship has the foreign key,
            # so that many inner elements can map to the same outer element.
            key_link = (relationship["parent_primary_key"], relationship["parent_foreign_key"])
        else:
            raise SyntaxError(f"Invalid relationship type: {relationship_type}")

    join_table_data = JoinAndSelectData(
        relations=relations,
        aliases=aliases,
        current_alias_str=current_alias_str,
        current_alias_sql_schema=current_alias_sql_schema,
        key_link=key_link,
        field_alias=db_field,
        search_properties=search_properties,
        unresolved=resolve[1:],
    )

    if len(resolve) == 1:
        # We're at the end of the resolve list - this is the last part of the 'drill-down' for the field
        return join_table_data,  # Return single tuple of relation

    if schema["type"] not in ("array", "object"):
        # Primitive type, len(resolve) > 1
        # TODO: Handle invalid schema types?
        raise TypeError("Cannot get property of primitive")

    if schema["type"] == "array" and resolve[1].value != "[item]":
        raise TypeError("Cannot get property of array in #resolve")
    elif schema["type"] == "object":  # Object
        if "properties" not in schema:
            raise SyntaxError("Searchable objects in schemas must have all properties described")
        if resolve[1].value not in schema["properties"]:
            raise ValueError(f"Property {resolve[1]} not found in object")

    return (join_table_data,) + collect_resolve_join_tables(
        resolve=resolve[1:],
        schema=schema["items"] if schema["type"] == "array" else schema["properties"][resolve[1].value],
        parent_relation=(relations.current, aliases.current),
        aliased_resolve_path=new_aliased_resolve_path if current_relation is not None else None,
    )


def collect_join_tables(ast: q.AST, terms: tuple, schema: JSONSchema) -> Tuple[JoinAndSelectData, ...]:
    if isinstance(ast, q.Literal):
        return terms

    if ast.fn == q.FUNCTION_RESOLVE:
        terms_list = list(terms)
        collected_joins = collect_resolve_join_tables((QUERY_ROOT, *ast.args), schema)

        for j in collected_joins:
            existing_aliases = set(t.current_alias_str for t in terms_list if t is not None)
            if j.current_alias_str is not None and j.current_alias_str not in existing_aliases:
                terms_list.append(j)

        return tuple(terms_list)

    new_terms = terms

    for item in (a for a in ast.args if isinstance(a, q.Expression)):
        new_terms = collect_join_tables(item, new_terms, schema)

    return new_terms


def join_fragment(ast: q.AST, schema: JSONSchema) -> sql.Composable:
    """
    When we query using our Bento search syntax, we don't have a concept of 'tables' in the relational DB sense.
    So, when we build the SELECT ... FROM ... WHERE query, we need to join a bunch of tables together, as specified
    in the search schema. This function builds that X JOIN Y JOIN Z ... fragment for the final select query.
    :param ast: The AST representation of the query being executed.
    :param schema: The JSON schema + extra Bento search properties for
    :return: The SQL fragment with all the aliased joins.
    """

    terms = collect_join_tables(ast, (), schema)
    if not terms:  # Query was probably just a literal
        # TODO: Don't hard-code _root?
        search_database_properties = _get_search_and_database_properties(schema)[1]
        relation = search_database_properties.get("relation")
        return sql.SQL("{r1} AS {a1}").format(
            r1=sql.Identifier(relation) if relation else sql.SQL("(SELECT NULL)"),
            a1=SQL_ROOT
        )

    return sql.SQL(", ").join((
        # Alias each table/relation to a re-usable name, joining the relations on their linked keys
        # specified in the search schema.

        # First, join any relation pairs which are connected together via foreign keys
        # (e.g., nested objects stored in their own relations).
        # If there is just 1 entry in terms, no join will occur, and it'll just be set to its alias.
        sql.SQL(" LEFT JOIN ").join((
            sql.SQL("{r1} AS {a1}{s1}").format(
                r1=terms[0].relations.current,
                a1=terms[0].aliases.current,
                s1=terms[0].current_alias_sql_schema or SQL_NOTHING,
            ),
            *(
                sql.SQL("{r1} AS {a1}{s1} ON {a0}.{f0} = {a1}.{f1}").format(
                    r1=term.relations.current,
                    a0=term.aliases.parent,
                    a1=term.aliases.current,
                    f0=sql.Identifier(term.key_link[0]),
                    f1=sql.Identifier(term.key_link[1]),
                    s1=term.current_alias_sql_schema or SQL_NOTHING,
                )
                for term in terms[1:]
                if term.key_link is not None
            ),  # Exclude terms without key-links - they will be selected separately below
        )),

        # Then, include any additional (non-terms[0]) non-joined relations.
        *(
            sql.SQL("{r1} AS {a1}{s1}").format(
                r1=term.relations.current,
                a1=term.aliases.current,
                s1=term.current_alias_sql_schema or SQL_NOTHING,
            )
            for term in terms[1:]
            if term.key_link is None and term.relations.current is not None
        ),
    ))


def search_ast_to_psycopg2_expr(ast: q.AST, params: tuple, schema: JSONSchema, internal: bool = False) \
        -> SQLComposableWithParams:
    if isinstance(ast, q.Literal):
        return sql.Placeholder(), (*params, ast.value)

    # Before doing anything, check that the permissions are correct given the AST and the search schema.
    #  TODO: use OIDC to maybe dynamically inject permissions/access levels somehow - either into schema or as a param
    q.check_operation_permissions(ast, schema, search_getter=get_search_properties, internal=internal)

    # Begin recursively constructing the SQL expression, starting with the top-most expression in our query
    return POSTGRES_SEARCH_LANGUAGE_FUNCTIONS[ast.fn](ast.args, params, schema, internal)


def search_query_to_psycopg2_sql(query, schema: JSONSchema, internal: bool = False) -> SQLComposableWithParams:
    # TODO: Shift recursion to not have to add in the extra SELECT for the root?
    ast = q.convert_query_to_ast_and_preprocess(query)
    sql_obj, params = search_ast_to_psycopg2_expr(ast, (), schema, internal)
    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    return sql.SQL("SELECT {root}.* FROM {relations_with_joins} WHERE {query_expr}").format(
        root=SQL_ROOT,
        relations_with_joins=join_fragment(ast, schema),
        query_expr=sql_obj
    ), params


def uncurried_binary_op(op: str, args: q.Args, params: tuple, schema: JSONSchema, internal: bool = False) \
        -> SQLComposableWithParams:
    # TODO: Need to fix params!! Use named params
    lhs_sql, lhs_params = search_ast_to_psycopg2_expr(args[0], params, schema, internal)
    rhs_sql, rhs_params = search_ast_to_psycopg2_expr(args[1], params, schema, internal)
    return (
        sql.SQL("({lhs}) {op} ({rhs})").format(lhs=lhs_sql, op=sql.SQL(op), rhs=rhs_sql),
        params + lhs_params + rhs_params,  # Collect all params together from elsewhere + both sides of the expression
    )


def _binary_op(op) -> Callable[[q.Args, tuple, JSONSchema, bool], SQLComposableWithParams]:
    return lambda args, params, schema, internal: uncurried_binary_op(op, args, params, schema, internal)


def _in(args: q.Args, params: tuple, schema: JSONSchema, internal: bool = False) -> SQLComposableWithParams:
    lhs_sql, lhs_params = search_ast_to_psycopg2_expr(args[0], params, schema, internal)
    rhs_sql, rhs_params = search_ast_to_psycopg2_expr(args[1], params, schema, internal)
    return sql.SQL("({lhs}) IN {rhs}").format(lhs=lhs_sql, rhs=rhs_sql), params + lhs_params + rhs_params


def _not(args: q.Args, params: tuple, schema: JSONSchema, internal: bool = False) -> SQLComposableWithParams:
    child_sql, child_params = search_ast_to_psycopg2_expr(args[0], params, schema, internal)
    return sql.SQL("NOT ({child})").format(child=child_sql), params + child_params


def _like_op(op: str) -> Callable[[Tuple[q.AST, q.AST], tuple, JSONSchema, bool], SQLComposableWithParams]:
    def inner(args: Tuple[q.AST, q.AST], params: tuple, schema: JSONSchema, internal: bool = False) \
            -> SQLComposableWithParams:
        lhs_sql, lhs_params = search_ast_to_psycopg2_expr(args[0], params, schema, internal)
        rhs_sql, rhs_params = search_ast_to_psycopg2_expr(args[1], params, schema, internal)

        return (
            sql.SQL("{lhs} {op} {rhs}").format(lhs=lhs_sql, op=sql.SQL(op), rhs=rhs_sql),
            params + lhs_params + rhs_params,
        )

    return inner


# TODO rename the function ?
def _contains(op: str, wc_loc: str, args: q.Args, params: tuple, schema: JSONSchema, internal: bool = False)\
        -> SQLComposableWithParams:
    return _like_op(op)(
        (args[0], q.Expression(fn=q.FUNCTION_HELPER_WC, args=[args[1], q.Literal(wc_loc)])), params, schema, internal)


def _contains_op(op: str) -> Callable[[q.Args, tuple, JSONSchema, bool], SQLComposableWithParams]:
    def _inner(*args):
        return _contains(op, "anywhere", *args)
    return _inner


_i_starts_with = functools.partial(_contains, "ILIKE", "start")
_i_ends_with = functools.partial(_contains, "ILIKE", "end")


def _wildcard(args: Tuple[q.AST, q.AST], params: tuple, _schema: JSONSchema, _internal: bool = False) \
        -> SQLComposableWithParams:
    if isinstance(args[0], q.Expression):
        raise NotImplementedError(f"Cannot currently use {q.FUNCTION_HELPER_WC} on an expression")  # TODO

    # TODO: py3.10: match
    if args[1].value == "start":
        wcs = "{}%"
    elif args[1].value == "end":
        wcs = "%{}"
    else:  # anywhere
        wcs = "%{}%"

    try:
        return sql.Placeholder(), (*params, wcs.format(args[0].value.replace("%", r"\%").replace("_", r"\_")))
    except AttributeError:
        # Can happen with non-string argument to #_wc, which will throw on .replace(...)
        raise TypeError(f"Type-invalid use of function {q.FUNCTION_HELPER_WC}")


def get_relation(resolve: Tuple[q.Literal, ...], schema: JSONSchema):
    aliases = collect_resolve_join_tables((QUERY_ROOT, *resolve), schema)[-1].aliases
    return aliases.current if aliases.current is not None else aliases.parent


def get_field(resolve: Tuple[q.Literal, ...], schema: JSONSchema) -> Optional[str]:
    return collect_resolve_join_tables((QUERY_ROOT, *resolve), schema)[-1].field_alias


def get_search_properties(resolve: Tuple[q.Literal, ...], schema: JSONSchema) -> dict:
    return collect_resolve_join_tables((QUERY_ROOT, *resolve), schema)[-1].search_properties


def _resolve(args: q.Args, params: tuple, schema: JSONSchema, _internal: bool = False) -> SQLComposableWithParams:
    """
    Compiles arguments for a #resolve call (an operation which accesses a field on a Bento searchable object)
    into SQL.
    :param args: Arguments representing the path to the field being resolved.
    :param params: Any existing SQL parameters (i.e, values to insert into the SQL without introducing injections).
    :param schema: The schema for the Bento searchable object.
    :param _internal: (unused here) whether we are querying from a global-access context, or a permissioned one.
    :return: A tuple of the SQL representation for the field access, and the params tuple (unchanged here).
    """
    f_id = get_field(args, schema)
    return sql.SQL("{relation}.{field}").format(
        relation=get_relation(args, schema),
        field=sql.Identifier(f_id) if f_id is not None else sql.SQL("*")
    ), params


def _list(args: q.Args, params: tuple, _schema: JSONSchema, _internal: bool = False) -> SQLComposableWithParams:
    # :param args: a tuple of query.Literal objects to be used in an IN clause.
    # with psycopg2, it must be passed as a tuple of tuples, hence the enclosing
    # parentheses in the following statement.
    # TODO: for large lists, the syntax IN (VALUES ("patient1"), ("patient2")) which creates
    #  a table construct on the fly is recommended performance-wise (better indexing)
    # FUTURE: psycopg3 won't allow the "tuples adaptation" syntax anymore:
    # https://www.psycopg.org/psycopg3/docs/basic/from_pg2.html#you-cannot-use-in-s-with-a-tuple
    values = tuple(a.value for a in args)
    return sql.Placeholder(), (*params, values)


POSTGRES_SEARCH_LANGUAGE_FUNCTIONS: Dict[
    str,
    Callable[[q.Args, tuple, JSONSchema, bool], SQLComposableWithParams]
] = {
    q.FUNCTION_AND: _binary_op("AND"),
    q.FUNCTION_OR: _binary_op("OR"),
    q.FUNCTION_NOT: _not,

    q.FUNCTION_LT: _binary_op("<"),
    q.FUNCTION_LE: _binary_op("<="),
    q.FUNCTION_EQ: _binary_op("="),
    q.FUNCTION_GT: _binary_op(">"),
    q.FUNCTION_GE: _binary_op(">="),
    q.FUNCTION_IN: _in,

    q.FUNCTION_CO: _contains_op("LIKE"),
    q.FUNCTION_ICO: _contains_op("ILIKE"),

    q.FUNCTION_ISW: _i_starts_with,
    q.FUNCTION_IEW: _i_ends_with,
    q.FUNCTION_LIKE: _like_op("LIKE"),
    q.FUNCTION_ILIKE: _like_op("ILIKE"),

    q.FUNCTION_RESOLVE: _resolve,
    q.FUNCTION_LIST: _list,

    q.FUNCTION_HELPER_WC: _wildcard,
}



__all__ = [
    "StreamingRangeNotSatisfiable",
    "StreamingBadRange",
    "StreamingProxyingError",
    "StreamingResponseExceededLimit",
    "StreamingBadURI",
    "StreamingUnsupportedURIScheme",
]


class StreamingRangeNotSatisfiable(Exception):
    def __init__(self, message: str, n_bytes: int | None):
        self._n_bytes: int | None = n_bytes
        super().__init__(message)

    @property
    def n_bytes(self) -> int:
        return self._n_bytes


class StreamingBadRange(Exception):
    pass


class StreamingProxyingError(Exception):
    pass


class StreamingResponseExceededLimit(Exception):
    pass


class StreamingBadURI(Exception):
    pass


class StreamingUnsupportedURIScheme(Exception):
    pass



import json
import re

import jsonschema

from functools import partial
from itertools import chain, product, starmap
from operator import and_, or_, not_, lt, le, eq, gt, ge, contains, is_not
from typing import Callable, Dict, List, Iterable, Optional, Tuple, Union

from . import queries as q
from ._types import JSONSchema


__all__ = ["check_ast_against_data_structure"]


QueryableStructure = Union[dict, list, set, str, int, float, bool]
BBOperator = Callable[[QueryableStructure, QueryableStructure], bool]

IndexCombination = Dict[str, int]
ArrayLengthData = Tuple[str, int, Tuple["ArrayLengthData", ...]]


def _icontains(lhs: str, rhs: str) -> bool:
    """
    Same as the "contains" operator, except with case-folded (i.e. case
    insensitive) arguments.
    :param lhs: The LHS for the operation.
    :param rhs: The RHS for the operation.
    :return: The result of the icontains operation.
    """
    return contains(lhs.casefold(), rhs.casefold())


def _in(lhs: Union[str, int, float], rhs: QueryableStructure) -> bool:
    """
    Same as `contains`, except order of arguments is inverted and second
    argument is a set.
    Note: set is preferred over list as it makes `contains` complexity O(1) vs O(n).
    As this is intended to be used in a loop over dataset results, it makes the
    implementation's complexity linear instead of quadratic.
    """
    return contains(rhs, lhs)


def _i_starts_with(lhs: str, rhs: str) -> bool:
    """
    Checks whether a string starts with a particular prefix, in a case-insensitive fashion.
    :param lhs: The full string to assess.
    :param rhs: The prefix to test against LHS.
    :return: Whether the string starts with the prefix.
    """
    if not isinstance(lhs, str) or not isinstance(rhs, str):
        raise TypeError(f"{q.FUNCTION_ISW} can only be used with strings")
    return lhs.casefold().startswith(rhs.casefold())


def _i_ends_with(lhs: str, rhs: str) -> bool:
    """
    Checks whether a string ends with a particular suffix, in a case-insensitive fashion.
    :param lhs: The full string to assess.
    :param rhs: The prefix to test against LHS.
    :return: Whether the string ends with the prefix.
    """
    if not isinstance(lhs, str) or not isinstance(rhs, str):
        raise TypeError(f"{q.FUNCTION_IEW} can only be used with strings")
    return lhs.casefold().endswith(rhs.casefold())


# See, e.g., https://stackoverflow.com/questions/399078/what-special-characters-must-be-escaped-in-regular-expressions
REGEX_CHARS_TO_ESCAPE = frozenset({"[", "]", "(", ")", "{", "}", "\\", ".", "^", "$", "*", "+", "-", "?", "|"})


def regex_from_like_pattern(pattern: str, case_insensitive: bool) -> re.Pattern:
    """
    Converts an SQL-style match pattern with %/_ wildcards into a Python Regex object.
    :param pattern: The SQL-style match pattern to convert.
    :param case_insensitive: Whether the generated Regex should be case-insensitive.
    :return: The converted Regex object.
    """

    # - Replace % with (.*) if % is not preceded by a \
    # - Wrap with ^$ to replicate whole-string behaviour
    # - Escape any special Regex characters

    regex_form: List[str] = ["^"]
    escape_mode: bool = False
    for char in pattern:
        # Put us into escape mode, so that the next character is escaped if needed
        if char == "\\":
            escape_mode = True
            continue

        if char == "%":  # Matches any number of characters
            # If we're in escape mode, append the literal %. Otherwise, replace it with a wildcard pattern.
            regex_form.append("%" if escape_mode else "(.*)")
        elif char == "_":  # Match a single character
            regex_form.append("_" if escape_mode else ".")
        elif char in REGEX_CHARS_TO_ESCAPE:
            # Escape special Regex characters with a backslash while building pattern
            regex_form.append(rf"\{char}")
        else:
            regex_form.append(char)  # Unmodified if not special

        escape_mode = False  # Turn off escape mode after one iteration; it only applies to the character in front of it

    regex_form.append("$")

    return re.compile("".join(regex_form), *((re.IGNORECASE,) if case_insensitive else ()))


def _like_op(case_insensitive: bool):
    def like_inner(lhs, rhs) -> bool:
        if not isinstance(lhs, str) or not isinstance(rhs, str):
            raise TypeError(f"{q.FUNCTION_LIKE} can only be used with strings")

        return regex_from_like_pattern(rhs, case_insensitive).match(lhs) is not None

    return like_inner


def _validate_data_structure_against_schema(
        data_structure: QueryableStructure, schema: JSONSchema, secure_errors: bool = True) -> None:
    """
    Validates a queryable data structure of some type against a JSON schema. This is an important validation step,
    because (assuming the schema is correct) it allows methods to make more assumptions about the integrity of the
    data structure while traversing it.
    :param data_structure: The data structure to validate
    :param schema: The JSON schema to validate the data structure against
    """
    schema_validator = jsonschema.Draft7Validator(schema)
    if not schema_validator.is_valid(data_structure):
        # There is a mismatch between the data structure and the corresponding
        # search schema. This probably means either the schema is incorrect or
        # the service is returning data that doesn't conform to what it says it
        # should conform to. If secure_errors is False, the data structure,
        # schema, and validation errors will all be returned in the giant
        # error string.

        errors = tuple(err.message for err in schema_validator.iter_errors(data_structure))
        errors_str = "\n".join(errors)

        if secure_errors:
            raise ValueError(f"Invalid data structure for schema (schema ID: {schema.get('$id', 'N/A')});"
                             f"encountered {len(errors)} validation errors (masked for privacy)")

        raise ValueError(
            f"Invalid data structure: \n"
            f"{data_structure}\n"
            f"For schema: \n"
            f"{json.dumps(schema)} \n"
            f"Validation Errors: \n"
            f"{errors_str}")


def _validate_not_wc(e: q.AST) -> None:
    """
    The #_wc (wildcard) expression function is a helper for converting the queries into the Postgres IR. If we encounter
    this function in a query being evaluated against a data structure, it's meaningless and should raise an error.
    :param e: The expression (function) to check
    """
    if e.fn == q.FUNCTION_HELPER_WC:
        raise NotImplementedError("Cannot use wildcard helper here")


def evaluate_no_validate(
    ast: q.AST,
    data_structure: QueryableStructure,
    schema: JSONSchema,
    index_combination: Optional[IndexCombination],
    internal: bool = False,
    resolve_checks: bool = True,
    check_permissions: bool = True,
) -> QueryableStructure:
    """
    Evaluates a query expression into a value, populated by a passed data structure.
    :param ast: A query expression.
    :param data_structure: A data structure from which to resolve values.
    :param schema: The JSON schema for data objects being queried.
    :param index_combination: The combination of array indices being evaluated upon.
    :param internal: Whether internal-only fields are allowed to be resolved.
    :param resolve_checks: Whether to run resolve checks. Should only be run once per query/ds/schema combo
    :param check_permissions: Whether to check the operation permissions. Typically called once per AST/DS combo.
    :return: A value (string, int, float, bool, array, or dict.)
    """

    # A literal (e.g. <Literal value=5>) evaluates to its own value (5)
    if ast.type == "l":
        return ast.value

    if resolve_checks:  # TODO: Separate setting
        # Prevents the Postgres internal-only #_wc function from being used in expressions being evaluated against
        # Python data structures. See the documentation for _validate_not_wc. Should only be run once per AST.
        _validate_not_wc(ast)

    if check_permissions:
        # Check that the current permissions (internal or not) allow us to perform the current operation on any resolved
        # fields. Internal queries are used for joins, etc. by services, or are performed by someone with unrestricted
        # access to the data.
        # TODO: This could be made more granular (some people could be given access to specific objects / tables)
        q.check_operation_permissions(
            ast,
            schema,
            lambda rl, s: _resolve_properties_and_check(rl, s, index_combination),
            internal)

    # Evaluate the non-literal expression recursively.
    return QUERY_CHECK_SWITCH[ast.fn](ast.args, data_structure, schema, index_combination, internal,
                                      resolve_checks, check_permissions)


def evaluate(
    ast: q.AST,
    data_structure: QueryableStructure,
    schema: JSONSchema,
    index_combination: Optional[IndexCombination],
    internal: bool = False,
    resolve_checks: bool = True,
    check_permissions: bool = True,
    secure_errors: bool = True,
):
    # The 'validate' flag is used to avoid redundantly validating the integrity of child data structures
    _validate_data_structure_against_schema(data_structure, schema, secure_errors=secure_errors)
    return evaluate_no_validate(ast, data_structure, schema, index_combination, internal, resolve_checks,
                                check_permissions)


def _collect_array_lengths(ast: q.AST, data_structure: QueryableStructure, schema: JSONSchema,
                           resolve_checks: bool) -> Iterable[ArrayLengthData]:
    """
    To evaluate a query in a manner consistent with the Postgres evaluator (and facilitate richer queries), each array
    item needs to be fixed in a particular evaluation of a query that involves array accesses. This helper function
    collects the lengths of arrays for each different array used in the field; it does this by traversing the data
    structure. These can be later used by _create_all_index_combinations to create all possible combinations of accesses
    to fix them in an evaluation run.
    :param ast: The AST-ified query
    :param data_structure: The FULL data structure the query is being evaluated against
    :param schema: The JSON schema of the full data structure
    :param resolve_checks: Whether to run resolve checks. Should only be run once per query/ds/schema combo
    :return: A recursive dictionary with keys being array paths and values being a tuple of (length, children dict)
    """

    # Literals are not arrays (currently), so they will not have any specified lengths
    if ast.type == "l":
        return ()

    # Standard validation to prevent Postgres internal-style queries from being passed in (see _validate_not_wc docs)
    _validate_not_wc(ast)

    # Resolves are where the magic happens w/r/t array access. Capture any array accesses with their lengths and child
    # array accesses.
    if ast.fn == q.FUNCTION_RESOLVE:
        r = _resolve_array_lengths(ast.args, data_structure, schema, "_root", resolve_checks)
        return () if r is None else (r,)

    # If the current expression is a non-resolve function, recurse into its arguments and collect any additional array
    # accesses; construct a list of possibly redundant array accesses with the arrays' lengths.
    als = tuple(chain.from_iterable(_collect_array_lengths(e, data_structure, schema, resolve_checks)
                                    for e in ast.args))
    return (
        a1 for i1, a1 in enumerate(als)
        if not any(
            a1[0] == a2[0] and len(a1[2]) <= len(a2[2])  # Deduplicate identical or subset items
            for a2 in als[i1+1:]
        )
    )


def _dict_combine(dicts: Iterable[dict]):
    """
    Utility function to combine an iterable of dictionaries into a single dictionary via d.update(d2)
    :param dicts: Iterable of dictionaries
    :return: A single, combined dictionary
    """
    c = {}
    for d in dicts:
        c.update(d)
    return c


def _create_index_combinations(parent_template: IndexCombination, array_data: ArrayLengthData) \
        -> Iterable[IndexCombination]:
    """
    Creates combinations of array indices from a particular array (including children, NOT including siblings.)
    :param parent_template: A dictionary with information about the array's parent's current fixed indexed configuration
    :param array_data: Information about an array's length and its children's lengths
    :return: An iterable of different combinations of fixed indices for the array and it's children (for later search)
    """

    # array_data is a tuple of (path, length, (tuple of child array lengths,))

    for i in range(array_data[1]):
        item_template = {**parent_template, array_data[0]: i}

        if len(array_data[2]) == 0:  # TODO: Don't like this logic...
            yield item_template  # What if this is overridden elsewhere?
            continue

        # TODO: Don't like the mono-tuple-ing stuff
        yield from _create_all_index_combinations(item_template, (array_data[2][i],))


def _create_all_index_combinations(parent_template: IndexCombination, arrays_data: Iterable[ArrayLengthData]) \
        -> Iterable[IndexCombination]:
    """
    Creates combinations of array indexes for all siblings in an iterable of arrays' length data.
    :param parent_template: A dictionary with information about the arrays' parent's current fixed indexed configuration
    :param arrays_data: An iterable of arrays' length data
    :return: An iterable of different combinations of fixed indices for the arrays and their children (for later search)
    """

    # Combine index mappings from different combination sets into a final list of array index combinations
    # Takes the cross product of the combination sets, since they're parallel fixations and there may be inter-item
    # comparisons between the two sets.
    # TODO: Do we need the combination superset replacement logic still?
    #  combinations = [c for c in combinations if not any(c.items() < c2.items() for c2 in combinations)]
    return map(
        _dict_combine,
        # Loop through and recurse
        product(*map(partial(_create_index_combinations, parent_template), arrays_data))
    )


# TODO: More rigorous / defined rules
def check_ast_against_data_structure(
    ast: q.AST,
    data_structure: QueryableStructure,
    schema: JSONSchema,
    internal: bool = False,
    return_all_index_combinations: bool = False,
    secure_errors: bool = True,
    skip_schema_validation: bool = False,
) -> Union[bool, Iterable[IndexCombination]]:
    """
    Checks a query against a data structure, returning True if the
    :param ast: A query to evaluate against the data object.
    :param data_structure: The data object to evaluate the query against.
    :param schema: A JSON schema representing valid data objects.
    :param internal: Whether internal-only fields are allowed to be resolved.
    :param return_all_index_combinations: Whether to return all index combinations that the query resolves to True on.
    :param secure_errors: Whether to not expose any data in error messaevaluateges. Impairs debugging.
    :param skip_schema_validation: Whether to skip schema validation on the data structure. Improves performance but can
                                   lead to wonky errors.
    :return: Determined by return_all_index_combinations; either
               1) A boolean representing whether or not the query matches the data object; or
               2) An iterable of all index combinations where the query matches the data object
    """

    if not skip_schema_validation:
        # Validate data structure against JSON schema here to avoid having to repetitively do it later
        _validate_data_structure_against_schema(data_structure, schema, secure_errors=secure_errors)

    # Collect all array resolves and their lengths in order to properly cross-product arrays
    array_lengths = _collect_array_lengths(ast, data_structure, schema, True)

    # Create all combinations of indexes into arrays and enumerate them; to be used to loop through all combinations of
    # array indices to freeze "[item]"s at particular indices across the whole query.
    index_combinations = enumerate(_create_all_index_combinations({}, array_lengths))

    # TODO: What to do here? Should be standardized, esp. w/r/t False returns

    def _evaluate(i: int, ic: IndexCombination) -> bool:
        return evaluate_no_validate(ast, data_structure, schema, ic, internal, False, (i == 0)) is True

    if return_all_index_combinations:
        return (ic for i, ic in index_combinations if _evaluate(i, ic))

    return any(starmap(_evaluate, index_combinations))


def _binary_op(op: BBOperator)\
        -> Callable[[q.Args, QueryableStructure, JSONSchema, Optional[IndexCombination], bool, bool, bool], bool]:
    """
    Returns a boolean-returning binary operator on a pair of arguments against a data structure/object of some type and
    return a Boolean result.
    :param op: The operator the lambda is representing
    :return: Operator lambda for use in evaluating expressions
    """

    # needed for shortcutting assessment of boolean operators
    is_and = op == and_
    is_or = op == or_

    def uncurried_binary_op(
        args: q.Args,
        ds: QueryableStructure,
        schema: JSONSchema,
        ic: Optional[IndexCombination],
        internal: bool,
        resolve_checks: bool,
        check_permissions: bool
    ) -> bool:
        # TODO: Standardize type safety / behaviour!!!

        # Evaluate both sides of the binary expression. If there's a type error while trying to use a Python built-in,
        # override it with a custom-message type error.

        lhs = evaluate_no_validate(args[0], ds, schema, ic, internal, resolve_checks, check_permissions)

        # These shortcuts mean that the RHS does NOT get type-checked!

        # Shortcut #and
        if is_and and not lhs:
            return False

        # Shortcut #or
        if is_or and lhs:
            return True

        rhs = evaluate_no_validate(args[1], ds, schema, ic, internal, resolve_checks, check_permissions)

        try:
            return op(lhs, rhs)
        except TypeError:
            raise TypeError(f"Type-invalid use of binary operator {op} ({lhs}, {rhs})")

    return uncurried_binary_op


def _resolve_checks(resolve_value: str, schema: JSONSchema):
    """
    Performs standard checks while going through any type of "resolve"-based function (where a #resolve call is being
    processed) to prevent access errors.
    :param resolve_value: The value of the current resolve term being processed
    :param schema: The JSON schema of the data sub-structure currently being processed
    """
    if schema["type"] not in ("object", "array"):
        raise TypeError("Cannot get property of literal")

    elif schema["type"] == "object" and resolve_value not in schema["properties"]:
        raise ValueError(f"Property {resolve_value} not found in object")

    elif schema["type"] == "array" and resolve_value != "[item]":
        raise TypeError("Cannot get property of array")


_is_not_none = partial(is_not, None)


def _get_child_resolve_array_lengths(
    new_resolve: Tuple[q.AST, ...],
    resolving_ds: list,
    item_schema: JSONSchema,
    new_path: str,
    resolve_checks: bool,
) -> Iterable[ArrayLengthData]:
    """
    Recursively resolve array lengths for all children of elements of an array using the _resolve_array_length function.
    :param new_resolve: The resolve path starting after the item access of the array being processed
    :param resolving_ds: The array data structure whose elements we're resolving child array accesses of
    :param item_schema: The JSON schema of the array's items
    :param new_path: The string representation of the path followed so far, including the most recent item access
    :param resolve_checks: Whether to run resolve checks. Should only be run once per query/ds/schema combo
    :return: A tuple of the current array's element-wise array length data
    """
    return filter(_is_not_none, (
        _resolve_array_lengths(new_resolve, array_item_ds, item_schema, new_path, resolve_checks)
        for array_item_ds in resolving_ds
    ))


def _resolve_array_lengths(
    resolve: Tuple[q.AST, ...],
    resolving_ds: QueryableStructure,
    schema: JSONSchema,
    path: str = "_root",
    resolve_checks: bool = True,
) -> Optional[ArrayLengthData]:
    """
    Given a resolve path and a data structure, find lengths of any arrays in the current data structure and any
    descendents it may have.
    :param resolve: The current resolve path, where the first element is the next thing to resolve on the data structure
    :param resolving_ds: The data structure we're resolving on
    :param schema: A JSON schema modeling the current data structure
    :param path: A string representation of the path followed so far, including the most recent access
    :param resolve_checks: Whether to run resolve checks. Should only be run once per query/ds/schema combo
    :return: Either none (if no arrays were accessed) or a tuple of the current array's path, its length, and the
             lengths of any child array accesses
    """

    # TODO: yield multiple for children instead of having child tuple?

    if len(resolve) == 0:
        # Resolve the root if it's an empty list
        #  - Python typing is awkward here (relying on schema correctness), so we don't process this line.
        return (path, len(resolving_ds), ()) if schema["type"] == "array" else None  # type: ignore

    resolve_value: str = str(resolve[0].value)

    if resolve_checks:  # pragma: no cover  TODO: Do we need this at all? right now we always check here
        _resolve_checks(resolve_value, schema)

    new_path = f"{path}.{resolve_value}"

    # The current data structure is an array, so return its length and recurse on its (potential) child arrays.
    if resolve[0].value == "[item]":
        return (
            path,
            len(resolving_ds),
            tuple(
                _get_child_resolve_array_lengths(resolve[1:], resolving_ds, schema["items"], new_path, resolve_checks)))

    # Otherwise, it's an object, so keep traversing without doing anything
    return _resolve_array_lengths(resolve[1:], resolving_ds[resolve_value], schema["properties"][resolve_value],
                                  new_path, resolve_checks)


def _resolve_properties_and_check(
    resolve: Tuple[q.Literal, ...],
    schema: JSONSchema,
    index_combination: Optional[IndexCombination],
) -> dict:
    """
    Resolves / evaluates a path (either object or array) into a value and its search properties. Assumes the
    data structure has already been checked against its schema.
    :param resolve: The current path to resolve, not including the current data structure
    :param schema: The JSON schema representing the resolving data structure
    :param index_combination: The combination of array indices being evaluated upon
    :return: The resolved value after exploring the resolve path, and the search operations that can be performed on it
    """

    path = "_root"
    r_schema = schema

    for current_resolve in resolve:
        current_resolve_value = current_resolve.value

        _resolve_checks(current_resolve_value, r_schema)
        if current_resolve_value == "[item]" and (index_combination is None or path not in index_combination):
            # TODO: Specific exception class
            raise Exception(f"Index combination not provided for path {path}")

        r_schema = r_schema["items"] if r_schema["type"] == "array" else r_schema["properties"][current_resolve_value]
        path = f"{path}.{current_resolve_value}"

    # Resolve the root if resolve list is empty
    return r_schema.get("search", {})


def _resolve(
    resolve: Tuple[q.Literal, ...],
    resolving_ds: QueryableStructure,
    _schema: JSONSchema,
    index_combination: Optional[IndexCombination],
    _internal: bool,
    _resolve_checks: bool,
    _check_permissions: bool,
) -> QueryableStructure:
    """
    Resolves / evaluates a path (either object or array) into a value. Assumes the data structure has already been
    checked against its schema.
    :param resolve: The current path to resolve, not including the current data structure
    :param resolving_ds: The data structure being resolved upon
    :param index_combination: The combination of array indices being evaluated upon
    :return: The resolved value after exploring the resolve path, and the search operations that can be performed on it
    """

    path = "_root"

    for current_resolve in resolve:
        current_resolve_value = current_resolve.value
        resolving_ds = (resolving_ds[index_combination[path]] if current_resolve_value == "[item]"
                        else resolving_ds[current_resolve_value])
        path = f"{path}.{current_resolve_value}"

    return resolving_ds


def _list(
    literals: Tuple[q.Literal, ...],
    _resolving_ds: QueryableStructure,
    _schema: JSONSchema,
    _index_combination: Optional[IndexCombination],
    _internal: bool,
    _resolve_checks: bool,
    _check_permissions: bool,
) -> QueryableStructure:
    """
    This function is to be used in conjonction with the #in operator to check
    for matches in a set of literals. (e.g. individual.karyotypic_sex in {"XX", "X0", "XXX"})
    :param literals: an iterable containing all the literals to check upon.
    :return: a set containing all the values
    """
    return set(literal.value for literal in literals)


QUERY_CHECK_SWITCH: Dict[
    q.FunctionName,
    Callable[[q.Args, QueryableStructure, JSONSchema, Optional[IndexCombination], bool, bool, bool],
             QueryableStructure]
] = {
    q.FUNCTION_AND: _binary_op(and_),
    q.FUNCTION_OR: _binary_op(or_),
    q.FUNCTION_NOT: lambda args, *rest: not_(evaluate_no_validate(args[0], *rest)),

    q.FUNCTION_LT: _binary_op(lt),
    q.FUNCTION_LE: _binary_op(le),
    q.FUNCTION_EQ: _binary_op(eq),
    q.FUNCTION_GT: _binary_op(gt),
    q.FUNCTION_GE: _binary_op(ge),

    q.FUNCTION_CO: _binary_op(contains),
    q.FUNCTION_ICO: _binary_op(_icontains),
    q.FUNCTION_IN: _binary_op(_in),

    q.FUNCTION_ISW: _binary_op(_i_starts_with),
    q.FUNCTION_IEW: _binary_op(_i_ends_with),
    q.FUNCTION_LIKE: _binary_op(_like_op(case_insensitive=False)),
    q.FUNCTION_ILIKE: _binary_op(_like_op(case_insensitive=True)),

    q.FUNCTION_RESOLVE: _resolve,
    q.FUNCTION_LIST: _list,
}



import re

from .exceptions import StreamingRangeNotSatisfiable, StreamingBadRange

__all__ = ["parse_range_header"]


BYTE_RANGE_INTERVAL_SPLIT = re.compile(r",\s*")
BYTE_RANGE_START_ONLY = re.compile(r"^(\d+)-$")
BYTE_RANGE_START_END = re.compile(r"^(\d+)-(\d+)$")
BYTE_RANGE_SUFFIX = re.compile(r"^-(\d+)$")


def parse_range_header(
    range_header: str | None, content_length: int, refget_mode: bool = False
) -> tuple[tuple[int, int], ...]:
    """
    Parse a range header (given a particular content length) into a validated series of sorted, non-overlapping
    start/end-inclusive intervals.
    """

    if range_header is None:
        return ((0, content_length - 1),)

    intervals: list[tuple[int, int]] = []

    if not range_header.startswith("bytes="):
        raise StreamingBadRange("only bytes range headers are supported")

    intervals_str = range_header.removeprefix("bytes=")

    # Cases: start- | start-end | -suffix, [start- | start-end | -suffix], ...

    intervals_str_split = BYTE_RANGE_INTERVAL_SPLIT.split(intervals_str)

    for iv in intervals_str_split:
        if m := BYTE_RANGE_START_ONLY.match(iv):
            intervals.append((int(m.group(1)), content_length - 1))
        elif m := BYTE_RANGE_START_END.match(iv):
            intervals.append((int(m.group(1)), int(m.group(2))))
        elif m := BYTE_RANGE_SUFFIX.match(iv):
            inclusive_content_length = content_length - 1
            suffix_length = int(m.group(1))  # suffix: -500 === last 500:
            intervals.append((max(inclusive_content_length - suffix_length + 1, 0), inclusive_content_length))
        else:
            raise StreamingBadRange("byte range did not match any pattern")

    intervals.sort()
    n_intervals: int = len(intervals)

    # validate intervals are not inverted and do not overlap each other:
    for i, int1 in enumerate(intervals):
        int1_start, int1_end = int1

        # Order of these checks is important - we want to give a 416 if start/end is beyond content length (which also
        # results in an inverted interval)

        if int1_start >= content_length:
            # both ends of the range are 0-indexed, inclusive - so it starts at 0 and ends at content_length - 1
            if refget_mode:  # sigh... GA4GH moment
                raise StreamingBadRange(f"start is beyond content length: {int1_start} >= {content_length}")
            raise StreamingRangeNotSatisfiable(f"not satisfiable: {int1_start} >= {content_length}", content_length)

        if int1_end >= content_length:
            # both ends of the range are 0-indexed, inclusive - so it starts at 0 and ends at content_length - 1
            if refget_mode:  # sigh... GA4GH moment
                raise StreamingBadRange(f"end is beyond content length: {int1_end} >= {content_length}")
            raise StreamingRangeNotSatisfiable(f"not satisfiable: {int1_end} >= {content_length}", content_length)

        if not refget_mode and int1_start > int1_end:
            raise StreamingRangeNotSatisfiable(f"inverted interval: {int1}", content_length)

        if i < n_intervals - 1:
            int2 = intervals[i + 1]
            int2_start, int2_end = int2

            if int1_end >= int2_start:
                raise StreamingRangeNotSatisfiable(f"intervals overlap: {int1}, {int2}", content_length)

    return tuple(intervals)



from . import notifications
from . import types
from ._event_bus import EventBus


__all__ = [
    "ALL_SERVICE_EVENTS",
    "ALL_DATA_TYPE_EVENTS",
    "EventBus",
    "notifications",
    "types",
]


ALL_SERVICE_EVENTS = "bento.service.*"
ALL_DATA_TYPE_EVENTS = "bento.data_type.*"



import logging

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError, StarletteHTTPException
from fastapi.middleware.cors import CORSMiddleware
from urllib.parse import urlparse

from bento_lib.auth.exceptions import BentoAuthException
from bento_lib.auth.middleware.fastapi import FastApiAuthMiddleware
from bento_lib.config.pydantic import BentoFastAPIBaseConfig
from bento_lib.responses.fastapi_errors import (
    bento_auth_exception_handler_factory,
    http_exception_handler_factory,
    validation_exception_handler_factory,
)
from bento_lib.service_info.helpers import build_service_info_from_pydantic_config
from bento_lib.service_info.types import BentoExtraServiceInfo, GA4GHServiceInfo, GA4GHServiceType

__all__ = ["BentoFastAPI"]


class BentoFastAPI(FastAPI):
    def __init__(
        self,
        authz_middleware: FastApiAuthMiddleware | None,
        config: BentoFastAPIBaseConfig,
        logger: logging.Logger,
        bento_extra_service_info: BentoExtraServiceInfo,
        service_type: GA4GHServiceType,
        version: str,
        exc_handler_kwargs: dict | None = None,
        *args,
        **kwargs,
    ):
        app_kwargs = dict(
            title=config.service_name,
            root_path=urlparse(config.service_url_base_path).path,
            docs_url=config.service_docs_path,
            openapi_url=config.service_openapi_path,
            version=version,
            **kwargs,
        )

        super().__init__(*args, **app_kwargs)

        self._config = config
        self._logger = logger
        self._bento_extra_service_info = bento_extra_service_info
        self._service_type = service_type
        self._version = version

        # Set up CORS
        self.add_middleware(
            CORSMiddleware,
            allow_origins=config.cors_origins,
            allow_credentials=True,
            allow_headers=["Authorization", "Cache-Control"],
            allow_methods=["*"],
        )

        # Set up authorization
        #  - Non-standard middleware setup so that we can import the instance and use it for dependencies too
        if authz_middleware:
            authz_middleware.attach(self)

        # Set up exception handlers for standard Bento/FastAPI errors
        exc_handler_kwargs = exc_handler_kwargs or {}
        self.exception_handler(BentoAuthException)(
            bento_auth_exception_handler_factory(logger, authz_middleware, **exc_handler_kwargs))
        self.exception_handler(StarletteHTTPException)(
            http_exception_handler_factory(logger, authz_middleware, **exc_handler_kwargs))
        self.exception_handler(RequestValidationError)(
            validation_exception_handler_factory(authz_middleware, **exc_handler_kwargs))

        # Set up service info endpoint

        self._service_info: GA4GHServiceInfo | None = None

        si_deps = [authz_middleware.dep_public_endpoint()] if authz_middleware else []

        @self.get("/service-info", dependencies=si_deps)
        async def service_info():
            return await self.get_service_info()

    async def get_service_info(self) -> GA4GHServiceInfo:
        if not self._service_info:
            self._service_info = await build_service_info_from_pydantic_config(
                self._config, self._logger, self._bento_extra_service_info, self._service_type, self._version)
        return self._service_info



__all__ = [
    "RESOURCE_EVERYTHING",
    "build_resource",
]


RESOURCE_EVERYTHING = {"everything": True}


def build_resource(project: str | None = None, dataset: str | None = None, data_type: str | None = None) -> dict:
    if project is None:
        return RESOURCE_EVERYTHING

    res = {"project": project}
    if dataset:
        res["dataset"] = dataset
    if data_type:
        res["data_type"] = data_type

    return res



import json
import jsonschema
import logging
import redis
import uuid

from datetime import datetime, timezone
from typing import Callable

__all__ = ["EventBus"]

# Channel templates
SERVICE_CHANNEL_TPL = "bento.service.{}"
DATA_TYPE_CHANNEL_TPL = "bento.data_type.{}"

# Types
Serializable = bool | float | int | str | dict | list | tuple | None

# Redis
default_connection_data = {"host": "localhost", "port": 6379}


class EventBus:
    """
    Event bus for subscribing to and publishing events for other microservices.
    """

    @staticmethod
    def _get_redis(**kwargs) -> redis.Redis:
        if "url" in kwargs:
            return redis.Redis.from_url(kwargs["url"])
        return redis.Redis(**kwargs)

    def __init__(self, allow_fake: bool = False, **kwargs):
        """
        Sets up a Redis connection based on the REDIS_SOCKET environment variable (or defaults, if the variable is
        not present.) Keyword arguments are passed to the Redis connection factory. If "url" is passed as a keyword
        argument, it is used instead of any other keyword arguments.
        :param allow_fake: Whether to allow for "fake" connections, i.e. no true connection to Redis.
        """

        self._rc: redis.Redis | None = None

        logger = kwargs.pop("logger", None) or logging.getLogger(__name__)
        self._logger: logging.Logger = logger

        connection_data: dict = kwargs or default_connection_data

        try:
            self._rc = self._get_redis(**connection_data)
            self._rc.get("")  # Dummy request to check connection
        except redis.exceptions.ConnectionError as e:
            self._rc = None
            if not allow_fake:
                raise e
            logger.warning(f"Starting event bus in 'fake' mode (tried connection data: {connection_data})")

        self._ps: redis.client.PubSub | None = None

        self._ps_handlers: dict[str, Callable[[dict], None]] = {}
        self._event_thread: redis.client.PubSubWorkerThread | None = None

        self._service_event_types: dict[str, dict] = {}
        self._data_type_event_types: dict[str, dict] = {}

    @staticmethod
    def _callback_deserialize(callback: Callable[[dict], None]) -> Callable[[dict], None]:
        return lambda message: callback({
            **message,
            "data": json.loads(message["data"]) if message["type"] in ("message", "pmessage") else message["data"]
        })

    def add_handler(self, pattern: str, callback: Callable[[dict], None]) -> bool:
        """
        Adds a channel pattern handler to the event bus if the event handling thread has not been started.
        :param pattern: Channel pattern (Redis syntax) to subscribe to.
        :param callback: Function to call (with message dictionary, see redis-py docs) when a matching event occurs.
        :return: True if the handler was successfully added, False otherwise.
        """

        if self._event_thread is not None:
            return False

        if pattern in self._ps_handlers:
            return False

        self._ps_handlers[pattern] = self._callback_deserialize(callback)
        return True

    def start_event_loop(self) -> None:
        """
        Starts the event handling loop in a new thread. Whichever handlers were previously added will be present.
        The loop must be restarted if the handlers are changed.
        """

        if self._rc is None:
            return

        if self._event_thread is not None:
            return

        self._logger.debug("Starting EventBus event loop")

        self._ps = self._rc.pubsub()

        if self._ps_handlers:  # Only try to subscribe if we have any registered handlers
            self._ps.psubscribe(**self._ps_handlers)

        self._event_thread = self._ps.run_in_thread(sleep_time=0.001, daemon=True)

    def stop_event_loop(self) -> None:
        """
        Stops the event handling loop, if running. Otherwise, does nothing.
        """

        if self._event_thread is None:
            return

        self._logger.debug("Stopping EventBus event loop")

        self._event_thread.stop()
        self._event_thread = None
        self._ps = None

    @staticmethod
    def _make_event(event_type: str, event_data, attrs: dict) -> str:
        return json.dumps({
            # Generate a random ID, so other services can decide how to claim specific events or whatever:
            "id": str(uuid.uuid4()),
            # Events can arrive out-of-order; we can put them back in order using this generation-time timestamp
            # (UTC timezone, in milliseconds):
            "timestamp": round(datetime.now(timezone.utc).timestamp() * 1000),
            # legacy version: "ts"  TODO: deprecated: remove
            "ts": datetime.now(timezone.utc).isoformat(),
            "type": event_type.lower(),
            "data": event_data,
            **attrs
        })

    @staticmethod
    def _add_schema(event_types: dict, event_type: str, event_schema: dict) -> bool:
        event_type = event_type.lower()

        if event_type in event_types:
            return False

        try:
            jsonschema.validators.Draft7Validator.check_schema(event_schema)
        except jsonschema.exceptions.SchemaError:
            return False

        event_types[event_type] = event_schema
        return True

    def register_service_event_type(self, event_type: str, event_schema: dict) -> bool:
        """
        Registers a service event type with the event bus.
        :param event_type: The type of event, which will be included in message objects.
        :param event_schema: JSON schema which will be checked against any events submitted.
        :return: True if the event type was successfully registered, False otherwise.
        """
        return self._add_schema(self._service_event_types, event_type, event_schema)

    def register_data_type_event_type(self, event_type: str, event_schema: dict) -> bool:
        """
        Registers a service event type with the event bus.
        :param event_type: The type of event, which will be included in message objects.
        :param event_schema: JSON schema which will be checked against any events submitted.
        :return: True if the event type was successfully registered, False otherwise.
        """
        return self._add_schema(self._data_type_event_types, event_type, event_schema)

    def get_service_event_types(self) -> dict[str, dict]:
        """
        :return: A dictionary of registered service event types and their associated schemas.
        """
        return {**self._service_event_types}

    def get_data_type_event_types(self) -> dict[str, dict]:
        """
        :return: A dictionary of registered data type event types and their associated schemas.
        """
        return {**self._data_type_event_types}

    def _publish_event(
        self,
        event_types: dict[str, dict],
        channel: str,
        event_type: str,
        event_data: Serializable,
        attrs: dict
    ) -> bool:
        event_type = event_type.lower()

        if event_type not in event_types:
            return False

        if not jsonschema.validators.Draft7Validator(event_types[event_type]).is_valid(event_data):
            return False

        if self._rc is None:
            return False

        event: str = self._make_event(event_type, event_data, attrs)
        self._logger.debug(f"Publishing event: {event}")
        self._rc.publish(channel, event)
        return True

    def publish_service_event(self, service_artifact: str, event_type: str, event_data: Serializable) -> bool:
        """
        Publishes a service event.
        :param service_artifact: Service artifact associated with the event to publish.
        :param event_type: The identifier for the event (not to be confused with its channel.)
        :param event_data: The data to send with the event; JSON-serializable.
        :return: True if the event was successfully published, False otherwise.
        """
        return self._publish_event(
            event_types=self._service_event_types,
            channel=SERVICE_CHANNEL_TPL.format(service_artifact),
            event_type=event_type,
            event_data=event_data,
            attrs={"service_artifact": service_artifact}
        )

    def publish_data_type_event(self, data_type: str, event_type: str, event_data: Serializable) -> bool:
        """
        Publishes a data type event.
        :param data_type: Data type associated with the event to publish.
        :param event_type: The identifier for the event (not to be confused with its channel.)
        :param event_data: The data to send with the event; JSON-serializable.
        :return: True if the event was successfully published, False otherwise.
        """
        return self._publish_event(
            event_types=self._data_type_event_types,
            channel=DATA_TYPE_CHANNEL_TPL.format(data_type),
            event_type=event_type,
            event_data=event_data,
            attrs={"data_type": data_type}
        )



import asyncio
import copy
import logging

from bento_lib.config.pydantic import BentoBaseConfig
from .constants import SERVICE_ENVIRONMENT_DEV, SERVICE_ENVIRONMENT_PROD, SERVICE_GROUP_BENTO
from .types import BentoExtraServiceInfo, GA4GHServiceType, GA4GHServiceOrganization, GA4GHServiceInfo


__all__ = [
    "build_service_info",
    "build_service_info_from_pydantic_config",
    "build_service_type",
    "build_bento_service_type",
]


async def _git_stdout(*args) -> str:
    git_proc = await asyncio.create_subprocess_exec(
        "git", *args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
    res, _ = await git_proc.communicate()
    return res.decode().rstrip()


async def build_service_info(
    base_service_info: GA4GHServiceInfo,
    debug: bool,
    local: bool,
    logger: logging.Logger,
) -> GA4GHServiceInfo:
    service_info_dict: GA4GHServiceInfo = copy.deepcopy(base_service_info)
    service_info_dict["environment"] = SERVICE_ENVIRONMENT_DEV if debug else SERVICE_ENVIRONMENT_PROD

    if not local:
        return service_info_dict

    try:
        res_tag, res_branch, res_commit = await asyncio.gather(
            _git_stdout("describe", "--tags", "--abbrev=0"),
            _git_stdout("branch", "--show-current"),
            _git_stdout("rev-parse", "HEAD"),
        )

        if "bento" not in service_info_dict:
            service_info_dict["bento"] = {}

        if res_tag:  # pragma: no cover
            # noinspection PyTypeChecker
            service_info_dict["bento"]["gitTag"] = res_tag
        if res_branch:  # pragma: no cover
            # noinspection PyTypeChecker
            service_info_dict["bento"]["gitBranch"] = res_branch
        if res_commit:  # pragma: no cover
            # noinspection PyTypeChecker
            service_info_dict["bento"]["gitCommit"] = res_commit

    except Exception as e:  # pragma: no cover
        except_name = type(e).__name__
        logger.error(f"Error retrieving git information: {str(except_name)}")

    return service_info_dict  # updated service info with the git info


async def build_service_info_from_pydantic_config(
    # dependencies
    config: BentoBaseConfig,
    logger: logging.Logger,
    # values for service info
    bento_service_info: BentoExtraServiceInfo,
    service_type: GA4GHServiceType,
    version: str,
) -> GA4GHServiceInfo:
    desc = config.service_description
    service_org: GA4GHServiceOrganization = config.service_organization.model_dump(mode="json")
    return await build_service_info({
        "id": config.service_id,
        "name": config.service_name,
        "type": service_type,
        **({"description": desc} if desc else {}),
        "organization": service_org,
        "contactUrl": config.service_contact_url,
        "version": version,
        "bento": bento_service_info,
    }, debug=config.bento_debug, local=config.bento_container_local, logger=logger)


def build_service_type(group: str, artifact: str, version: str) -> GA4GHServiceType:
    return {
        "group": group,
        "artifact": artifact,
        "version": version,
    }


def build_bento_service_type(artifact: str, version: str) -> GA4GHServiceType:
    return build_service_type(SERVICE_GROUP_BENTO, artifact, version)



from pydantic import BaseModel, ConfigDict
from typing import Literal, TypedDict

__all__ = [
    "GA4GHServiceType",
    "GA4GHServiceOrganization",
    "GA4GHServiceOrganizationModel",
    "BentoExtraServiceInfo",
    "GA4GHServiceInfo",
]


class GA4GHServiceType(TypedDict):
    group: str
    artifact: str
    version: str


class GA4GHServiceOrganization(TypedDict):
    name: str
    url: str


class GA4GHServiceOrganizationModel(BaseModel):
    name: str
    url: str
    # frozen=True makes this hashable + immutable
    model_config = ConfigDict(extra="forbid", frozen=True)


# TODO: py3.11: Required[] instead of base class


class BentoExtraServiceInfo(TypedDict, total=False):
    serviceKind: str  # One service_kind per Bento service/instance
    dataService: bool  # Whether the service provides data types/search/workflows
    # Whether the service provides workflows:
    #   - not necessarily data types; split from dataService to allow services to provide workflows
    #     without needing to provide data types/search endpoints as well
    #   - implict default: false
    workflowProvider: bool
    # Git:
    #  - Static Git information: should be set by service developers
    gitRepository: str
    #  - Dynamic Git information: only added if we're in a local development mode
    gitTag: str
    gitBranch: str
    gitCommit: str


class _GA4GHServiceInfoBase(TypedDict):
    id: str
    name: str
    type: GA4GHServiceType
    organization: GA4GHServiceOrganization
    version: str


class GA4GHServiceInfo(_GA4GHServiceInfoBase, total=False):
    description: str
    contactUrl: str
    documentationUrl: str
    url: str  # Technically not part of spec; comes from service-registry
    environment: Literal["dev", "prod"]
    # Bento-specific service info properties are contained inside a nested, "bento"-keyed dictionary
    bento: BentoExtraServiceInfo



import aiofiles
import asyncpg
import contextlib
from pathlib import Path
from typing import AsyncIterator


__all__ = [
    "PgAsyncDatabase",
]


class PgAsyncDatabase:

    def __init__(self, db_uri: str, schema_path: Path):
        self._db_uri: str = db_uri
        self._schema_path: Path = schema_path

        self._pool: asyncpg.Pool | None = None

    async def initialize(self, pool_size: int = 10):
        conn: asyncpg.Connection

        if not self._pool:  # Initialize the connection pool if needed
            self._pool = await asyncpg.create_pool(self._db_uri, min_size=pool_size, max_size=pool_size)

        # Connect to the database and execute the schema script
        async with aiofiles.open(self._schema_path, "r") as sf:
            async with self.connect() as conn:
                async with conn.transaction():
                    await conn.execute(await sf.read())

    async def close(self):
        if self._pool:
            await self._pool.close()
            self._pool = None

    @contextlib.asynccontextmanager
    async def connect(self, existing_conn: asyncpg.Connection | None = None) -> AsyncIterator[asyncpg.Connection]:
        # TODO: raise raise DatabaseError("Pool is not available") when FastAPI has lifespan dependencies
        #  + manage pool lifespan in lifespan fn.

        if self._pool is None:
            await self.initialize()  # initialize if this is the first time we're using the pool

        if existing_conn is not None:
            yield existing_conn
            return

        conn: asyncpg.Connection
        async with self._pool.acquire() as conn:
            yield conn



__all__ = [
    "BentoAuthException",
]


class BentoAuthException(Exception):
    def __init__(self, message="Unauthorized", status_code=401):
        super().__init__(message)
        self.message = message
        self.status_code = status_code



import aiohttp
import logging
import re
import requests

from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable

from bento_lib.config.pydantic import BentoBaseConfig
from ..exceptions import BentoAuthException
from ..permissions import Permission
from ..types import EvaluationResultMatrix, EvaluationResultDict
from .mark_authz_done_mixin import MarkAuthzDoneMixin

__all__ = ["BaseAuthMiddleware"]


def _compile_to_regex_if_needed(pattern: str | re.Pattern) -> re.Pattern:
    if isinstance(pattern, str):
        return re.compile(pattern)
    return pattern


class BaseAuthMiddleware(ABC, MarkAuthzDoneMixin):
    def __init__(
        self,
        bento_authz_service_url: str,
        drs_compat: bool = False,
        sr_compat: bool = False,
        beacon_meta_callback: Callable[[], dict] | None = None,
        exempt_request_patterns: tuple[tuple[re.Pattern | str, re.Pattern | str], ...] = (),
        debug_mode: bool = False,
        enabled: bool = True,
        logger: logging.Logger | None = None,
    ):
        self._debug: bool = debug_mode
        self._verify_ssl: bool = not debug_mode

        self._enabled: bool = enabled
        self._logger: logging.Logger | None = logger

        self._drs_compat: bool = drs_compat
        self._sr_compat: bool = sr_compat
        self._beacon_meta_callback: Callable[[], dict] | None = beacon_meta_callback

        self._exempt_request_patterns: frozenset[tuple[re.Pattern, re.Pattern]] = frozenset(
            (_compile_to_regex_if_needed(method_pattern), _compile_to_regex_if_needed(path_pattern))
            for method_pattern, path_pattern in exempt_request_patterns
        )

        self._bento_authz_service_url: str = bento_authz_service_url

    @classmethod
    def build_from_pydantic_config(cls, config: BentoBaseConfig, logger: logging.Logger, **kwargs):
        return cls(
            bento_authz_service_url=config.bento_authz_service_url,
            debug_mode=config.bento_debug,
            enabled=config.bento_authz_enabled,
            logger=logger,
            **kwargs,
        )

    @property
    def enabled(self) -> bool:
        return self._enabled

    def request_is_exempt(self, method: str, path: str) -> bool:
        return method == "OPTIONS" or any(
            mp.fullmatch(method) and pp.fullmatch(path) for mp, pp in self._exempt_request_patterns)

    @abstractmethod
    def get_authz_header_value(self, request: Any) -> str | None:  # pragma: no cover
        pass

    @staticmethod
    def check_require_token(require_token: bool, token: str | None) -> None:
        if require_token:
            if token is None:
                raise BentoAuthException("No token provided")

    def mk_authz_url(self, path: str) -> str:
        return f"{self._bento_authz_service_url.rstrip('/')}{path}"

    def _extract_token_and_build_headers(
        self,
        request: Any,
        require_token: bool,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict[str, str]:
        if headers_getter:
            return headers_getter(request)

        tkn_header = self.get_authz_header_value(request)
        self.check_require_token(require_token, tkn_header)
        return {"Authorization": tkn_header} if tkn_header else {}

    def _log_error(self, message: str):
        if self._logger:
            self._logger.error(message)

    def _gen_exc_non_200_error_from_authz(self, code: int, content: bytes):
        self._log_error(f"Got non-200 response from authorization service: {code} {content!r}")
        # Generic error - don't leak errors from authz service!
        raise BentoAuthException("Error from authz service", status_code=500)

    def authz_post(
        self,
        request: Any,
        path: str,
        body: dict,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict:
        res = requests.post(
            self.mk_authz_url(path),
            json=body,
            headers=self._extract_token_and_build_headers(request, require_token, headers_getter),
            verify=self._verify_ssl)

        if res.status_code != 200:  # Invalid authorization service response
            raise self._gen_exc_non_200_error_from_authz(res.status_code, res.content)

        return res.json()

    @staticmethod
    def _evaluate_body(resources: Iterable[dict], permissions: Iterable[Permission]) -> dict:
        return {"resources": tuple(resources), "permissions": tuple(permissions)}

    @staticmethod
    def _matrix_tuple_cast(authz_result: list[list[bool]]) -> EvaluationResultMatrix:
        return tuple(map(tuple, authz_result))

    @staticmethod
    def _permissions_matrix_to_dict(
        m: EvaluationResultMatrix,
        permissions: Iterable[Permission],
    ) -> EvaluationResultDict:
        return tuple(dict(zip(permissions, r)) for r in m)

    def evaluate(
        self,
        request: Any,
        resources: Iterable[dict],
        permissions: Iterable[Permission],
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> EvaluationResultMatrix:
        if mark_authz_done:
            self.mark_authz_done(request)
        return self._matrix_tuple_cast(
            self.authz_post(
                request,
                "/policy/evaluate",
                self._evaluate_body(resources, permissions),
                require_token=require_token,
                headers_getter=headers_getter,
            )["result"]
        )

    def evaluate_to_dict(
        self,
        request: Any,
        resources: Iterable[dict],
        permissions: Iterable[Permission],
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> EvaluationResultDict:
        # consume iterable only once in case it's a generator
        _perms = tuple(permissions)
        return self._permissions_matrix_to_dict(
            self.evaluate(request, resources, _perms, require_token, headers_getter, mark_authz_done), _perms)

    def evaluate_one(
        self,
        request: Any,
        resource: dict,
        permission: Permission,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> bool:
        return self.evaluate(request, (resource,), (permission,), require_token, headers_getter, mark_authz_done)[0][0]

    async def async_authz_post(
        self,
        request: Any,
        path: str,
        body: dict,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                    self.mk_authz_url(path),
                    json=body,
                    headers=self._extract_token_and_build_headers(request, require_token, headers_getter),
                    ssl=(None if self._verify_ssl else False)) as res:

                if res.status != 200:  # Invalid authorization service response
                    raise self._gen_exc_non_200_error_from_authz(res.status, await res.content.read())

                return await res.json()

    async def async_evaluate(
        self,
        request: Any,
        resources: Iterable[dict],
        permissions: Iterable[Permission],
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> EvaluationResultMatrix:
        if mark_authz_done:
            self.mark_authz_done(request)
        return self._matrix_tuple_cast(
            (
                await self.async_authz_post(
                    request,
                    "/policy/evaluate",
                    self._evaluate_body(resources, permissions),
                    require_token=require_token,
                    headers_getter=headers_getter,
                )
            )["result"]
        )

    async def async_evaluate_to_dict(
        self,
        request: Any,
        resources: Iterable[dict],
        permissions: Iterable[Permission],
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> EvaluationResultDict:
        # consume iterable only once in case it's a generator
        _perms = tuple(permissions)
        return self._permissions_matrix_to_dict(
            await self.async_evaluate(request, resources, _perms, require_token, headers_getter, mark_authz_done),
            _perms)

    async def async_evaluate_one(
        self,
        request: Any,
        resource: dict,
        permission: Permission,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
        mark_authz_done: bool = False,
    ) -> bool:
        return (
            await self.async_evaluate(
                request, (resource,), (permission,), require_token, headers_getter, mark_authz_done
            )
        )[0][0]

    def check_authz_evaluate(
        self,
        request: Any,
        permissions: frozenset[Permission],
        resource: dict,
        require_token: bool = True,
        set_authz_flag: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> None:
        if not self.enabled:
            return

        res = self.evaluate(
            request,
            [resource],
            list(permissions),
            require_token=require_token,
            headers_getter=headers_getter,
            mark_authz_done=set_authz_flag,
        )[0]

        if not all(res):
            # We early-return with the flag set - we're returning Forbidden,
            # and we've determined authz, so we can just set the flag.
            raise BentoAuthException("Forbidden", status_code=403)  # Actually forbidden by authz service

    async def async_check_authz_evaluate(
        self,
        request: Any,
        permissions: frozenset[Permission],
        resource: dict,
        require_token: bool = True,
        set_authz_flag: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> None:
        if not self.enabled:
            return

        res = (
            await self.async_evaluate(
                request,
                [resource],
                list(permissions),
                require_token=require_token,
                headers_getter=headers_getter,
                mark_authz_done=set_authz_flag,
            )
        )[0]

        if not all(res):
            # We early-return with the flag set - we're returning Forbidden,
            # and we've determined authz, so we can just set the flag.
            raise BentoAuthException("Forbidden", status_code=403)  # Actually forbidden by authz service



from fastapi import status
from fastapi.exceptions import HTTPException
from fastapi.responses import FileResponse
from fastapi.routing import APIRouter

from ..auth.middleware.fastapi import FastApiAuthMiddleware
from .models import WorkflowDefinition
from .workflow_set import WorkflowSet

__all__ = [
    "build_workflow_router",
]


def build_workflow_router(authz_middleware: FastApiAuthMiddleware, workflow_set: WorkflowSet) -> APIRouter:
    workflow_router = APIRouter(prefix="/workflows")

    @workflow_router.get("", dependencies=[authz_middleware.dep_public_endpoint()])
    def workflow_list():
        return workflow_set.workflow_dicts_by_type_and_id()

    @workflow_router.get("/", dependencies=[authz_middleware.dep_public_endpoint()])
    def workflow_list_trailing_slash():
        return workflow_set.workflow_dicts_by_type_and_id()

    # The endpoint with the .wdl suffix needs to come first, since we match in order and otherwise the whole thing,
    # including the suffix, would get passed as {workflow_id}.
    @workflow_router.get("/{workflow_id}.wdl", dependencies=[authz_middleware.dep_public_endpoint()])
    def workflow_file(workflow_id: str):
        if (wdl := workflow_set.get_workflow_wdl_path(workflow_id)) is not None:
            return FileResponse(wdl)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"No workflow with ID {workflow_id}")

    @workflow_router.get("/{workflow_id}", dependencies=[authz_middleware.dep_public_endpoint()])
    def workflow_item(workflow_id: str) -> WorkflowDefinition:
        if (wf := workflow_set.get_workflow(workflow_id)) is not None:
            return wf
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"No workflow with ID {workflow_id}")

    return workflow_router



from importlib import metadata

from . import apps
from . import auth
from . import drs
from . import events
from . import schemas
from . import search
from . import service_info
from . import streaming
from . import workflows

__version__ = metadata.version(__name__)
__all__ = [
    "__version__", "apps", "auth", "drs", "events", "schemas", "search", "service_info", "streaming", "workflows"
]



import aiohttp
import requests
import sys

from urllib.parse import urlparse

from typing import Optional

__all__ = [
    "DrsInvalidScheme",
    "DrsRequestError",

    "get_access_method_of_type",
    "decode_drs_uri",
    "fetch_drs_record_by_uri"
]


class DrsInvalidScheme(Exception):
    pass


class DrsRequestError(Exception):
    pass


def get_access_method_of_type(drs_object_record: dict, access_type: str, default=None):
    """
    Gets an access method of specified type from a DRS object, if one with the type exists.
    :param drs_object_record: The record for the DRS object, fetched from the instance.
    :param access_type: The access type to try and find in the object record.
    :param default: The value to return if the access method is not found.
    :return: The access method of the specified type for the object's bytes, if one exists.
    """
    return next((a for a in drs_object_record.get("access_methods", []) if a.get("type", None) == access_type), default)


def decode_drs_uri(drs_uri: str, internal_drs_base_url: Optional[str] = None) -> str:
    """
    Given a DRS URI and possibly an override for the DRS service URL, returns
    the decoded HTTP URL for the DRS object.
    :param drs_uri: The drs://-schemed URI for the DRS object.
    :param internal_drs_base_url: An optional override hard-coded DRS base URL to use, for container networking etc.
    :return: The HTTP URI for the DRS object specified.
    """

    parsed_drs_uri = urlparse(drs_uri)

    if parsed_drs_uri.scheme != "drs":
        print(f"[Bento Lib] Invalid scheme: '{parsed_drs_uri.scheme}'",
              file=sys.stderr, flush=True)
        raise DrsInvalidScheme(f"Encountered invalid DRS scheme: {parsed_drs_uri.scheme}")

    drs_base_path = internal_drs_base_url.rstrip("/") if internal_drs_base_url else f"https://{parsed_drs_uri.netloc}"
    return f"{drs_base_path}/ga4gh/drs/v1/objects/{parsed_drs_uri.path.split('/')[-1]}"


def fetch_drs_record_by_uri(drs_uri: str, internal_drs_base_url: Optional[str] = None) -> Optional[dict]:
    """
    Given a URI in the format drs://<hostname>/<object-id>, decodes it into an
    HTTP URL and fetches the object metadata.
    :param drs_uri: The URI of the object to fetch.
    :param internal_drs_base_url: An optional override hard-coded DRS base URL to use, for container networking etc.
    :return: The fetched DRS object metadata.
    """

    # TODO: Translation dictionary for internal DRS hostnames, to avoid overriding EVERY DRS host.

    decoded_object_uri = decode_drs_uri(drs_uri, internal_drs_base_url)
    print(f"[Bento Lib] Attempting to fetch {decoded_object_uri}", flush=True)
    params = {"internal_path": "true"} if internal_drs_base_url else {}
    drs_res = requests.get(decoded_object_uri, params=params)

    if drs_res.status_code != 200:
        print(f"[Bento Lib] Could not fetch: '{decoded_object_uri}'", file=sys.stderr, flush=True)
        print(f"\tAttempted URL: {decoded_object_uri} (status: {drs_res.status_code})", file=sys.stderr, flush=True)
        raise DrsRequestError(f"Could not fetch '{decoded_object_uri}' (status: {drs_res.status_code})")

    # TODO: Handle JSON parse errors
    # TODO: Schema for DRS response

    return drs_res.json()


async def fetch_drs_record_by_uri_async(drs_uri: str, internal_drs_base_url: Optional[str] = None) -> Optional[dict]:
    """
    Given a URI in the format drs://<hostname>/<object-id>, decodes it into an
    HTTP URL and asynchronously fetches the object metadata.
    :param drs_uri: The URI of the object to fetch.
    :param internal_drs_base_url: An optional override hard-coded DRS base URL to use, for container networking etc.
    :return: The fetched DRS object metadata.
    """

    # TODO: Translation dictionary for internal DRS hostnames, to avoid overriding EVERY DRS host.

    decoded_object_uri = decode_drs_uri(drs_uri, internal_drs_base_url)
    print(f"[Bento Lib] Attempting to fetch {decoded_object_uri}", flush=True)

    params = {"internal_path": "true"} if internal_drs_base_url else {}
    async with aiohttp.ClientSession() as session:
        async with session.get(decoded_object_uri, params=params) as drs_res:
            if drs_res.status != 200:
                print(f"[Bento Lib] Could not fetch: '{decoded_object_uri}'", file=sys.stderr, flush=True)
                print(f"\tAttempted URL: {decoded_object_uri} (status: {drs_res.status})", file=sys.stderr, flush=True)
                raise DrsRequestError(f"Could not fetch '{decoded_object_uri}' (status: {drs_res.status})")

            # TODO: Handle JSON parse errors
            # TODO: Schema for DRS response

            return await drs_res.json()



# !!! LEGACY FILE !!!

from django.contrib.auth.backends import RemoteUserBackend
from django.contrib.auth.middleware import RemoteUserMiddleware
from rest_framework.authentication import RemoteUserAuthentication

from bento_lib.auth.headers import DJANGO_USER_HEADER, DJANGO_USER_ROLE_HEADER
from bento_lib.auth.roles import ROLE_OWNER, ROLE_USER


__all__ = [
    "BentoRemoteUserAuthentication",
    "BentoRemoteUserBackend",
    "BentoRemoteUserMiddleware",
]


class BentoRemoteUserAuthentication(RemoteUserAuthentication):
    header = DJANGO_USER_HEADER


class BentoRemoteUserMiddleware(RemoteUserMiddleware):
    header = DJANGO_USER_HEADER


class BentoRemoteUserBackend(RemoteUserBackend):
    # noinspection PyMethodMayBeStatic
    def configure_user(self, request, user):
        is_owner = request.META.get(DJANGO_USER_ROLE_HEADER, ROLE_USER) == ROLE_OWNER
        user.is_staff = is_owner
        user.is_superuser = is_owner
        user.save()
        return user



import werkzeug.utils
from collections import defaultdict
from pathlib import Path
from .models import WorkflowDefinition

__all__ = [
    "WorkflowSet",
]


class WorkflowSet:
    """
    A class for constructing a singleton object that stores all workflow descriptions in a particular service.
    """

    def __init__(self, wdl_directory: Path):
        self._defs_by_id: dict[str, WorkflowDefinition] = {}
        self._wdl_directory: Path = wdl_directory

    def add_workflow(self, id_: str, definition: WorkflowDefinition):
        if id_ in self._defs_by_id:
            raise ValueError(f"Workflow with id {id_} already exists")

        self._defs_by_id[id_] = definition

    def get_workflow(self, id_: str) -> WorkflowDefinition | None:
        return self._defs_by_id.get(id_)

    def get_workflow_resource(self, id_: str) -> str | None:
        if (wd := self.get_workflow(id_)) is not None:
            return werkzeug.utils.secure_filename(wd.file)
        return None

    def get_workflow_wdl_path(self, id_: str) -> Path | None:
        if (wfr := self.get_workflow_resource(id_)) is not None:
            return self._wdl_directory / wfr
        return None

    def workflow_exists(self, id_: str) -> bool:
        return id_ in self._defs_by_id

    def workflow_dicts_by_type_and_id(self) -> dict[str, dict[str, dict]]:
        by_type = defaultdict(dict)
        for id_, wf in self._defs_by_id.items():
            by_type[wf.type][id_] = wf.model_dump(mode="json")
        return dict(by_type)

    def workflow_dicts_by_id(self) -> dict[str, WorkflowDefinition]:
        return {k: v.model_dump(mode="json") for k, v in self._defs_by_id.items()}



# TODO: Python 3.9: lowercase dict typing
from typing import Dict

__all__ = ["format_notification"]


def format_notification(title: str, description: str, notification_type: str, action_target: str) -> Dict[str, str]:
    """
    Takes in all parameters needed to construct a notification and returns a
    dictionary in the correct format.
    :param title: Notification title
    :param description: Notification description
    :param notification_type: Notification type
    :param action_target: Parameter for the target of any interaction with the notification (e.g. a click)
    :return: The formatted notification dictionary
    """
    return {
        "title": title,
        "description": description,
        "notification_type": notification_type,
        "action_target": action_target
    }



from typing import Type

from ..auth.middleware.mark_authz_done_mixin import MarkAuthzDoneMixin
from ..auth.permissions import Permission

__all__ = [
    "EvaluationResultMatrix",
    "EvaluationResultDict",
    "MarkAuthzDoneType",
]

EvaluationResultMatrix = tuple[tuple[bool, ...], ...]
EvaluationResultDict = tuple[dict[Permission, bool], ...]

# Allow subclass OR instance, since mark_authz_done is a static method:
MarkAuthzDoneType = MarkAuthzDoneMixin | Type[MarkAuthzDoneMixin]



import logging

from fastapi import Depends, FastAPI, Request, Response, status
from fastapi.responses import JSONResponse
from typing import Awaitable, Callable

from bento_lib.auth.exceptions import BentoAuthException
from bento_lib.auth.middleware.base import BaseAuthMiddleware
from bento_lib.auth.permissions import Permission
from bento_lib.auth.resources import RESOURCE_EVERYTHING
from bento_lib.responses.errors import http_error

__all__ = [
    "FastApiAuthMiddleware",
]


class FastApiAuthMiddleware(BaseAuthMiddleware):
    def attach(self, app: FastAPI):
        """
        Attach the middleware to an application. Must be called in order for requests to be checked.
        :param app: A FastAPI application.
        """

        # Attach our instance's dispatch method to the FastAPI instance as a middleware function
        app.middleware("http")(self.dispatch)

        # If no logger was passed, create a new logger
        if self._logger is None:
            self._logger = logging.getLogger(__name__)

    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        if not self.enabled or self.request_is_exempt(request.method, request.url.path):
            # - Skip checks if the authorization middleware is disabled
            # - Allow pre-flight responses through, as well as any configured exempt URLs
            return await call_next(request)

        # Set flag saying the request hasn't had its permissions determined yet.
        request.state.bento_determined_authz = False

        try:
            res: Response = await call_next(request)
            if not request.state.bento_determined_authz:
                # Next in response chain didn't properly think about auth; return 403
                raise BentoAuthException(status_code=status.HTTP_403_FORBIDDEN, message="Forbidden")

        except BentoAuthException as e:
            self.mark_authz_done(request)
            return JSONResponse(
                status_code=e.status_code,
                content=http_error(
                    e.status_code,
                    e.message,
                    drs_compat=self._drs_compat,
                    sr_compat=self._sr_compat,
                    beacon_meta_callback=self._beacon_meta_callback,
                ))

        # Otherwise, return the response as normal
        return res

    def get_authz_header_value(self, request: Request) -> str | None:
        return request.headers.get("Authorization")

    @staticmethod
    def mark_authz_done(request: Request):
        request.state.bento_determined_authz = True

    def dep_public_endpoint(self):
        def _inner(request: Request):
            if not self.enabled:
                return
            self.mark_authz_done(request)
        return Depends(_inner)

    def dep_require_permissions_on_resource(
        self,
        permissions: frozenset[Permission],
        resource: dict | None = None,
        require_token: bool = True,
        set_authz_flag: bool = True,
    ):
        resource = resource or RESOURCE_EVERYTHING

        async def _inner(request: Request):
            if not self.enabled:
                return

            await self.async_check_authz_evaluate(
                request,
                permissions,
                resource,
                require_token=require_token,
                set_authz_flag=set_authz_flag,
            )

        return Depends(_inner)



EVENT_WES_RUN_UPDATED = "wes_run_updated"
EVENT_WES_RUN_UPDATED_SCHEMA = {
    "type": "object",
    "properties": {
        "run_id": {"type": "string"},
        "request": {"type": "object"},  # TODO
        "state": {"type": "string"},  # TODO: Enum
        "run_log": {"type": "object"},  # TODO
        "task_logs": {"type": "array"},  # TODO
        "outputs": {"type": "object"},  # TODO
    },
    # TODO: deduplicate this schema
}

EVENT_WES_RUN_FINISHED = "wes_run_finished"
EVENT_WES_RUN_FINISHED_SCHEMA = {
    "type": "object",
    # TODO
}

EVENT_CREATE_NOTIFICATION = "create_notification"
EVENT_CREATE_NOTIFICATION_SCHEMA = {
    "type": "object",
    "properties": {
        "title": {"type": "string"},
        "description": {"type": "string"},
        "notification_type": {"type": "string"},
        "action_target": {"type": "string"},
    },
    "required": ["title", "description", "notification_type", "action_target"]
}

EVENT_NOTIFICATION = "notification"
EVENT_NOTIFICATION_SCHEMA = {
    "type": "object",
    "properties": {
        "id": {"type": "string"},
        "title": {"type": "string"},
        "description": {"type": "string"},
        "notification_type": {"type": "string"},
        "action_target": {"type": "string"},
        "read": {"type": "boolean"},
        "timestamp": {"type": "string"}
    },
    "required": ["id", "title", "description", "notification_type", "action_target", "read", "timestamp"]
}



import logging
import sys
import traceback

from flask import jsonify, request
from functools import partial
from typing import Callable

from ..auth.types import MarkAuthzDoneType
from ..responses import errors


__all__ = [
    "flask_error_wrap_with_traceback",
    "flask_error_wrap",

    "flask_error",

    "flask_bad_request_error",
    "flask_unauthorized_error",
    "flask_forbidden_error",
    "flask_not_found_error",
    "flask_method_not_allowed_error",
    "flask_not_acceptable_error",
    "flask_request_timeout_error",
    "flask_range_not_satisfiable_error",

    "flask_internal_server_error",
    "flask_not_implemented_error",
]


# noinspection PyIncorrectDocstring
def flask_error_wrap_with_traceback(fn: Callable, *args, **kwargs) -> Callable:
    """
    Function to wrap flask_* error creators with something that supports the application.register_error_handler method,
    while also printing a traceback. Optionally, the keyword argument service_name can be passed in to make the error
    logging more precise.
    :param fn: The flask error-generating function to wrap
    :return: The wrapped function
    """

    service_name = kwargs.pop("service_name", "Bento Service")

    logger: logging.Logger | None = kwargs.pop("logger", None)
    authz: MarkAuthzDoneType | None = kwargs.pop("authz", None)

    def handle_error(e):
        if logger:
            logger.error(f"Encountered error:\n{traceback.format_exception(type(e), e, e.__traceback__)}")
        else:
            print(f"[{service_name}] Encountered error:", file=sys.stderr)
            # TODO: py3.10: print_exception(e)
            traceback.print_exception(type(e), e, e.__traceback__)
        if authz:
            authz.mark_authz_done(request)
        return fn(str(e), *args, **kwargs)

    return handle_error


def flask_error_wrap(fn: Callable, *args, **kwargs) -> Callable:
    """
    Function to wrap flask_* error creators with something that supports the application.register_error_handler method
    and pass in the exception as a message.
    :param fn: The flask error-generating function to wrap
    :return: The wrapped function
    """

    authz: MarkAuthzDoneType | None = kwargs.pop("authz", None)

    def handle_error(e):
        if authz:
            authz.mark_authz_done(request)
        return fn(str(e), *args, **kwargs)

    return handle_error


def flask_error(code: int, *errs, **kwargs):
    return jsonify(errors.http_error(code, *errs, **kwargs)), code


def _flask_error(code: int) -> Callable:
    return partial(flask_error, code)


flask_bad_request_error = _flask_error(400)
flask_unauthorized_error = _flask_error(401)
flask_forbidden_error = _flask_error(403)
flask_not_found_error = _flask_error(404)
flask_method_not_allowed_error = _flask_error(405)
flask_not_acceptable_error = _flask_error(406)
flask_request_timeout_error = _flask_error(408)
flask_range_not_satisfiable_error = _flask_error(416)

flask_internal_server_error = _flask_error(500)
flask_not_implemented_error = _flask_error(501)



import logging
import traceback

from fastapi import status
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.requests import Request
from fastapi.responses import JSONResponse
from starlette.responses import Response
from typing import Callable

from ..auth.exceptions import BentoAuthException
from ..auth.types import MarkAuthzDoneType
from .errors import http_error

__all__ = [
    "http_exception_handler_factory",
    "bento_auth_exception_handler_factory",
    "validation_exception_handler_factory",
]


def _log_if_500(logger: logging.Logger, code: int, exc: Exception) -> None:
    if code == status.HTTP_500_INTERNAL_SERVER_ERROR:
        logger.error(f"Encountered error:\n{traceback.format_exception(type(exc), exc, exc.__traceback__)}")


def http_exception_handler_factory(
    logger: logging.Logger,
    authz: MarkAuthzDoneType | None = None,
    **kwargs,
) -> Callable[[Request, HTTPException], Response]:
    def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
        if authz:
            authz.mark_authz_done(request)
        code = exc.status_code
        _log_if_500(logger, code, exc)
        return JSONResponse(http_error(code, exc.detail, **kwargs), status_code=code)

    return http_exception_handler


def bento_auth_exception_handler_factory(
    logger: logging.Logger,
    authz: MarkAuthzDoneType | None = None,
    **kwargs,
) -> Callable[[Request, BentoAuthException], Response]:
    def bento_auth_exception_handler(request: Request, exc: BentoAuthException) -> JSONResponse:
        if authz:
            authz.mark_authz_done(request)
        code = exc.status_code
        _log_if_500(logger, code, exc)
        return JSONResponse(http_error(code, exc.message, **kwargs), status_code=code)

    return bento_auth_exception_handler


def validation_exception_handler_factory(
    authz: MarkAuthzDoneType | None = None,
    **kwargs,
) -> Callable[[Request, RequestValidationError], Response]:
    def validation_exception_handler(request: Request, exc: RequestValidationError) -> JSONResponse:
        if authz:
            authz.mark_authz_done(request)
        code = status.HTTP_400_BAD_REQUEST
        return JSONResponse(
            http_error(
                code,
                *((".".join(map(str, e["loc"])) + ": " + e["msg"])
                  if e.get("loc") else e["msg"] for e in exc.errors()),
                **kwargs),
            status_code=code)
    return validation_exception_handler



from pydantic import BaseModel, ConfigDict, field_serializer
from typing import Literal


__all__ = [
    # Input base modles
    "WorkflowBaseInput",
    "WorkflowInjectedInput",
    # Input models
    "WorkflowStringInput",
    "WorkflowStringInput",
    "WorkflowStringArrayInput",
    "WorkflowNumberInput",
    "WorkflowNumberArrayInput",
    "WorkflowBooleanInput",
    "WorkflowEnumInput",
    "WorkflowEnumArrayInput",
    "WorkflowProjectDatasetInput",
    "WorkflowFileInput",
    "WorkflowFileArrayInput",
    "WorkflowDirectoryInput",
    "WorkflowDirectoryArrayInput",
    "WorkflowServiceUrlInput",
    "WorkflowConfigInput",
    "WorkflowSecretInput",
    # Workflow definition model
    "WorkflowDefinition",
]


class FrozenBaseModel(BaseModel):
    model_config = ConfigDict(frozen=True)


class WorkflowBaseInput(FrozenBaseModel):
    id: str
    required: bool = True
    type: str
    help: str = ""  # Help text to render alongside fields in GUI forms.


class WorkflowInjectedInput(WorkflowBaseInput):
    injected: Literal[True] = True


class WorkflowStringInput(WorkflowBaseInput):
    type: Literal["string"] = "string"


class WorkflowStringArrayInput(WorkflowBaseInput):
    type: Literal["string[]"] = "string[]"


class WorkflowNumberInput(WorkflowBaseInput):
    type: Literal["number"] = "number"


class WorkflowNumberArrayInput(WorkflowBaseInput):
    type: Literal["number[]"] = "number[]"


class WorkflowBooleanInput(WorkflowBaseInput):
    type: Literal["boolean"] = "boolean"


class WorkflowEnumInput(WorkflowBaseInput):
    type: Literal["enum"] = "enum"
    values: list[str] | str  # list of values, or a URL returning an array of enum values


class WorkflowEnumArrayInput(WorkflowBaseInput):
    type: Literal["enum[]"] = "enum[]"
    values: list[str] | str  # list of values, or a URL returning an array of enum values
    repeatable: bool = True


class WorkflowProjectDatasetInput(WorkflowBaseInput):
    type: Literal["project:dataset"] = "project:dataset"


class WorkflowFileInput(WorkflowBaseInput):
    # can be sourced from drop box / DRS / workflow outputs, whatever the UI decides works
    type: Literal["file"] = "file"
    pattern: str = "*"  # file name regular expression pattern


class WorkflowFileArrayInput(WorkflowBaseInput):
    # can be sourced from drop box / DRS / workflow outputs, whatever the UI decides works
    type: Literal["file[]"] = "file[]"
    pattern: str = "*"  # file name regular expression pattern


class WorkflowDirectoryInput(WorkflowBaseInput):
    # can be sourced from drop box / DRS / workflow outputs, whatever the UI decides works
    type: Literal["directory"] = "directory"


class WorkflowDirectoryArrayInput(WorkflowBaseInput):
    # can be sourced from drop box / DRS / workflow outputs, whatever the UI decides works
    type: Literal["directory[]"] = "directory[]"


class WorkflowServiceUrlInput(WorkflowInjectedInput):
    # service URL from the service registry, using bento.serviceKind as a lookup
    type: Literal["service-url"] = "service-url"
    service_kind: str


class WorkflowConfigInput(WorkflowInjectedInput):
    # configuration injection from the workflow executor - stored in the database
    type: Literal["config"] = "config"
    key: str


class WorkflowSecretInput(WorkflowInjectedInput):
    # secret injection from the workflow executor - not present in the database, passed ephemerally
    type: Literal["secret"] = "secret"
    key: str


WorkflowInput = (
    WorkflowStringInput |
    WorkflowStringArrayInput |
    WorkflowNumberInput |
    WorkflowNumberArrayInput |
    WorkflowBooleanInput |
    WorkflowEnumInput |
    WorkflowEnumArrayInput |
    WorkflowProjectDatasetInput |
    WorkflowFileInput |
    WorkflowFileArrayInput |
    WorkflowDirectoryInput |
    WorkflowDirectoryArrayInput |
    WorkflowServiceUrlInput |
    WorkflowConfigInput |
    WorkflowSecretInput
)


WorkflowType = Literal["ingestion", "analysis", "export"]


class WorkflowDefinition(BaseModel):
    """
    Class defining meta-information about a workflow in the context of a Bento node.
    """
    name: str  # Human-readable workflow name
    type: WorkflowType  # One of a few pre-defined values for categorizing workflow type/purpose
    description: str  # Human-readable workflow description
    file: str  # WDL file name
    data_type: str | None = None  # Data type; temporary for authz until we have proper token exchange for WES  TODO
    #  - If data_type is None, the permissions should be *more* severe, not less
    #    (check if they have whole project/dataset access)
    tags: frozenset[str] = frozenset()  # Should include data type(s) if relevant
    # Here, inputs defines UI / injected inputs for this workflow. These get transformed into a JSON parameters file
    # which is fed to the WDL workflow description / Cromwell.
    # As such, many of these workflow input types end up mapping to the same WDL type:
    #  - ex. the Bento WorkflowInput types enum/project:datraset/dataset/ref-genome all map to the WDL type String.
    inputs: list[WorkflowInput]

    @field_serializer("tags")
    def serialize_permissions(self, tags: frozenset[str], _info):
        # make set serialization have a consistent order
        return sorted(tags)



import json

from flask import Flask, Request, Response, request, g
from functools import wraps

from bento_lib.auth.exceptions import BentoAuthException
from bento_lib.auth.middleware.base import BaseAuthMiddleware
from bento_lib.auth.permissions import Permission
from bento_lib.auth.resources import RESOURCE_EVERYTHING
from bento_lib.responses.errors import http_error

__all__ = [
    "FlaskAuthMiddleware",
]


class FlaskAuthMiddleware(BaseAuthMiddleware):
    def attach(self, app: Flask):
        """
        Attach the middleware to an application. Must be called in order for requests to be checked.
        :param app: A Flask application.
        """

        app.before_request(self.middleware_pre)
        app.after_request(self.middleware_post)

        if self._logger is None:
            self._logger = app.logger

    def middleware_pre(self) -> None:
        if self.enabled:
            g.bento_determined_authz = False

    def _make_auth_error(self, e: BentoAuthException) -> Response:
        # returning an error, so mark authz flow as 'done' (rejecting in one way or another):
        self.mark_authz_done(request)
        # return error response:
        return Response(
            json.dumps(http_error(
                e.status_code,
                e.message,
                drs_compat=self._drs_compat,
                sr_compat=self._sr_compat,
                beacon_meta_callback=self._beacon_meta_callback,
            )),
            status=e.status_code,
            content_type="application/json")

    def _make_forbidden(self) -> Response:
        return self._make_auth_error(BentoAuthException("Forbidden", status_code=403))

    def middleware_post(self, response: Response) -> Response:
        if not self.enabled or self.request_is_exempt(request.method, request.path):
            # - Skip checks if the authorization middleware is disabled
            # - Allow pre-flight responses through, as well as any configured exempt URLs
            return response
        if not g.bento_determined_authz:
            return self._make_forbidden()
        return response

    @staticmethod
    def mark_authz_done(_request: Request):
        g.bento_determined_authz = True

    def get_authz_header_value(self, r: Request) -> str | None:
        return r.headers.get("Authorization")

    # TODO: IMPL SYNC + ASYNC + TEST
    # def get_permissions_on_resource(self, resource: dict):
    #     return self.authz_post(
    #         request,
    #         "/policy/list-permissions",
    #         {"requested_resource": resource},
    #         require_token=False,
    #     ).get("result")

    def deco_require_permissions_on_resource(
        self,
        permissions: frozenset[Permission],
        resource: dict | None = None,
        require_token: bool = True,
        set_authz_flag: bool = True,
    ):
        resource = resource or RESOURCE_EVERYTHING  # If no resource specified, require the permissions node-wide.

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if self.enabled:
                    try:
                        self.check_authz_evaluate(request, permissions, resource, require_token, set_authz_flag)
                    except BentoAuthException as e:
                        return self._make_auth_error(e)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def deco_public_endpoint(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            if self.enabled:
                self.mark_authz_done(request)
            return res
        return wrapper



from typing import Iterable, NewType


class PermissionDefinitionError(Exception):
    pass


Level = NewType("Level", str)

LEVEL_DATASET = Level("dataset")
LEVEL_PROJECT = Level("project")
LEVEL_INSTANCE = Level("instance")

# Levels (applicability): Specifies the minimum (most specific) level at which a permission is applicable
#  - e.g., create:project is not a relevant permission at the project or dataset level, just at the instance level
# In order of most specific to most general:
LEVELS = [
    LEVEL_DATASET,
    LEVEL_PROJECT,
    LEVEL_INSTANCE,
]

PermissionVerb = NewType("PermissionVerb", str)
PermissionNoun = NewType("PermissionNoun", str)


PERMISSIONS: list["Permission"] = []
PERMISSIONS_BY_STRING: dict[str, "Permission"] = {}


class Permission(str):
    def __init__(
        self, verb: PermissionVerb,
        noun: PermissionNoun,
        min_level_required: Level = LEVEL_DATASET,
        supports_data_type_narrowing: bool | None = None,
        gives: Iterable["Permission"] = (),
    ):
        super().__init__()

        self._verb: PermissionVerb = verb
        self._noun: PermissionNoun = noun
        self._min_level_required: Level = min_level_required
        self._supports_data_type_narrowing: bool = (
            supports_data_type_narrowing if supports_data_type_narrowing is not None
            else min_level_required != LEVEL_INSTANCE
        )

        # Create full set using nested gives
        base_gives = set(gives)  # in case our iterable is a generator, consume it only once
        full_gives: set["Permission"] = set(base_gives)  # copy base_gives as a starting point
        for g in base_gives:
            full_gives |= g.gives
        self._gives: frozenset["Permission"] = frozenset(full_gives)

        str_rep = str(self)

        if str_rep in PERMISSIONS_BY_STRING:
            raise PermissionDefinitionError(f"Permission {str_rep} already defined")

        PERMISSIONS.append(self)
        PERMISSIONS_BY_STRING[str_rep] = self

    def __new__(
        cls,
        verb: PermissionVerb,
        noun: PermissionNoun,
        min_level_required: Level = LEVEL_DATASET,
        supports_data_type_narrowing: bool | None = None,
        gives: Iterable["Permission"] = (),
    ):
        return super().__new__(cls, cls._str_form(verb, noun))

    @classmethod
    def _str_form(cls, verb: PermissionVerb, noun: PermissionNoun) -> str:
        return f"{verb}:{noun}"

    def __repr__(self):
        return f"Permission({str(self)})"

    @property
    def verb(self) -> PermissionVerb:
        return self._verb

    @property
    def noun(self) -> PermissionNoun:
        return self._noun

    @property
    def gives(self) -> frozenset["Permission"]:
        return self._gives

    @property
    def min_level_required(self) -> Level:
        return self._min_level_required

    @property
    def supports_data_type_narrowing(self) -> bool:
        return self._supports_data_type_narrowing


# Verb/noun definitions ---------------------------------------------------------------------------

QUERY_VERB = PermissionVerb("query")
DOWNLOAD_VERB = PermissionVerb("download")
VIEW_VERB = PermissionVerb("view")
CREATE_VERB = PermissionVerb("create")
EDIT_VERB = PermissionVerb("edit")
DELETE_VERB = PermissionVerb("delete")
INGEST_VERB = PermissionVerb("ingest")
ANALYZE_VERB = PermissionVerb("analyze")
EXPORT_VERB = PermissionVerb("export")

PRIVATE_PORTAL = PermissionNoun("private_portal")

PROJECT_LEVEL_BOOLEAN = PermissionNoun("project_level_boolean")
DATASET_LEVEL_BOOLEAN = PermissionNoun("dataset_level_boolean")

PROJECT_LEVEL_COUNTS = PermissionNoun("project_level_counts")
DATASET_LEVEL_COUNTS = PermissionNoun("dataset_level_counts")

DATA = PermissionNoun("data")
DROP_BOX = PermissionNoun("drop_box")
RUNS = PermissionNoun("runs")

PROJECT = PermissionNoun("project")
DATASET = PermissionNoun("dataset")

#  - Other non-project/dataset-attached items:
NOTIFICATIONS = PermissionNoun("notifications")
REFERENCE_MATERIAL = PermissionNoun("reference_material")
PERMISSIONS_NOUN = PermissionNoun("permissions")

# Permissions definitions -------------------------------------------------------------------------

P_VIEW_PRIVATE_PORTAL = Permission(VIEW_VERB, PRIVATE_PORTAL, min_level_required=LEVEL_INSTANCE)

# TODO: embedding project/dataset here is awkward

P_QUERY_PROJECT_LEVEL_BOOLEAN = Permission(QUERY_VERB, PROJECT_LEVEL_BOOLEAN, min_level_required=LEVEL_PROJECT)
P_QUERY_DATASET_LEVEL_BOOLEAN = Permission(QUERY_VERB, DATASET_LEVEL_BOOLEAN)

P_QUERY_PROJECT_LEVEL_COUNTS = Permission(
    QUERY_VERB, PROJECT_LEVEL_COUNTS, min_level_required=LEVEL_PROJECT, gives=(P_QUERY_PROJECT_LEVEL_BOOLEAN,))
P_QUERY_DATASET_LEVEL_COUNTS = Permission(QUERY_VERB, DATASET_LEVEL_COUNTS, gives=(P_QUERY_DATASET_LEVEL_BOOLEAN,))

# Data-level: interacting with data inside of data services... and triggering workflows

P_QUERY_DATA = Permission(
    QUERY_VERB, DATA, gives=(P_QUERY_PROJECT_LEVEL_COUNTS, P_QUERY_DATASET_LEVEL_COUNTS))  # query at full access
P_DOWNLOAD_DATA = Permission(DOWNLOAD_VERB, DATA)  # download CSVs, associated DRS objects
P_DELETE_DATA = Permission(DELETE_VERB, DATA)  # clear data from a specific data type

#  - workflow-relevant items: (types of workflows....)

P_INGEST_DATA = Permission(INGEST_VERB, DATA)
P_ANALYZE_DATA = Permission(ANALYZE_VERB, DATA)
P_EXPORT_DATA = Permission(EXPORT_VERB, DATA)

P_VIEW_RUNS = Permission(VIEW_VERB, RUNS)

#  - notifications  TODO: notifications should have a resource embedded in them

P_VIEW_NOTIFICATIONS = Permission(VIEW_VERB, NOTIFICATIONS)
P_CREATE_NOTIFICATIONS = Permission(CREATE_VERB, NOTIFICATIONS)

#  - dataset metadata editing
P_EDIT_DATASET = Permission(EDIT_VERB, DATASET, supports_data_type_narrowing=False)

#  - can view edit permissions for the resource which granted this permission only:
P_VIEW_PERMISSIONS = Permission(VIEW_VERB, PERMISSIONS_NOUN)
P_EDIT_PERMISSIONS = Permission(EDIT_VERB, PERMISSIONS_NOUN)

# ---

# only {everything: true} or {project: ...} (instance- or project-level):
#  - project metadata editing
#  - dataset management
P_EDIT_PROJECT = Permission(EDIT_VERB, PROJECT, min_level_required=LEVEL_PROJECT, supports_data_type_narrowing=False)
P_CREATE_DATASET = Permission(
    CREATE_VERB, DATASET, min_level_required=LEVEL_PROJECT, supports_data_type_narrowing=False)
#     - deleting a dataset inherently deletes data inside it, so we give delete:data to all holders of delete:dataset
P_DELETE_DATASET = Permission(
    DELETE_VERB, DATASET, min_level_required=LEVEL_PROJECT, supports_data_type_narrowing=False, gives=(P_DELETE_DATA,))
# ---

# only {everything: true} (instance-level):

#  - drop box

P_VIEW_DROP_BOX = Permission(VIEW_VERB, DROP_BOX, min_level_required=LEVEL_INSTANCE)
P_INGEST_DROP_BOX = Permission(INGEST_VERB, DROP_BOX, min_level_required=LEVEL_INSTANCE)
P_DELETE_DROP_BOX = Permission(DELETE_VERB, DROP_BOX, min_level_required=LEVEL_INSTANCE)

#  - project management

P_CREATE_PROJECT = Permission(CREATE_VERB, PROJECT, min_level_required=LEVEL_INSTANCE)
#     - deleting a project inherently deletes datasets/data inside it,
#       so we give delete:data and delete:dataset to all holders of delete:project
P_DELETE_PROJECT = Permission(
    DELETE_VERB, PROJECT, min_level_required=LEVEL_INSTANCE, gives=(P_DELETE_DATASET, P_DELETE_DATA))

#  - reference material

#     * ingest gives create+edit permissions for genomes and annotations
P_INGEST_REFERENCE_MATERIAL = Permission(INGEST_VERB, REFERENCE_MATERIAL, min_level_required=LEVEL_INSTANCE)
P_DELETE_REFERENCE_MATERIAL = Permission(DELETE_VERB, REFERENCE_MATERIAL, min_level_required=LEVEL_INSTANCE)
#     * could add a P_VIEW_REFERENCE_MATERIAL which we by default give to everyone including anonymous...
